//****************************************************************************//
// File:          af_canpie.c                                                 //
// Description:                                                               //
// Author:        Uwe Koppe                                                   //
// e-mail:        koppe@microcontrol.net                                      //
//                                                                            //
//============================================================================//
// This program is free software; you can redistribute it and/or modify       //
// it under the terms of the GNU Lesser General Public License as published   //
// by the Free Software Foundation; either version 2.1 of the License, or     //
// (at your option) any later version.                                        //
//============================================================================//
//                                                                            //
// Date        History                                                        //
// ----------  -------------------------------------------------------------- //
// 16.02.2004  Initial version                                                //
//                                                                            //
//****************************************************************************//




//------------------------------------------------------------------------------
// CVS version information:
// $Id: af_can.c,v 1.4 2005/07/01 14:12:37 microcontrol Exp $
//------------------------------------------------------------------------------


/*----------------------------------------------------------------------------*\
** Include files                                                              **
**                                                                            **
\*----------------------------------------------------------------------------*/

#include <linux/config.h>
#include <linux/module.h>

#include <net/sock.h>

#include <net/can/af_can.h>
#include <net/can/canpie_dbg.h>
#include <net/can/canpie_hal.h>
#include <net/can/cpcanmsg.h>
#include <net/can/cpcore.h>

// #include <linux/can.h>              // CAN definitions


//-----------------------------------------------------------------------------
/*!   \file    af_can.c
**
**
*/


/*----------------------------------------------------------------------------*\
** Function implementations                                                   **
**                                                                            **
\*----------------------------------------------------------------------------*/
static struct proto_ops can_raw_ops;



//----------------------------------------------------------------------------//
// can_sock_alloc()                                                           //
// create a CAN sock structure                                                //
//----------------------------------------------------------------------------//
static int can_sock_alloc(struct sock * sk)
{
   struct can_sock * ptsCanSockT;
 

   //-----------------------------------------------------------------
   // create a new 'can_sock' structure and zero its contents
   //
   ptsCanSockT = kmalloc(sizeof(struct can_sock), GFP_ATOMIC);
   if(ptsCanSockT == NULL)
   {
      return(-ENOMEM);
   }
   
   memset(ptsCanSockT, 0x00, sizeof(struct can_sock));
   
   //-----------------------------------------------------------------
   // now link the 'sock' and the 'can_sock' together
   //
   ptsCanSockT->ptsSock       = sk;
   ptsCanSockT->rcv_fifo_size = CP_QUEUE_MSG_DEF;
   ptsCanSockT->trm_fifo_size = CP_QUEUE_MSG_DEF;
   
   sk->sk_protinfo = ptsCanSockT;
   
   
   PK_DBG(eDBG_SOCK, "created new CAN socket (%p)", ptsCanSockT);
   return(0);
}


//----------------------------------------------------------------------------//
// can_sock_release()                                                         //
// remove a CAN sock structure                                                //
//----------------------------------------------------------------------------//
static void can_sock_release(struct sock *sk)
{
   struct can_sock *csock;

   csock = sk->sk_protinfo;
   if(csock)
   {
      PK_DBG(4, "can_sock is %p", csock);
      kfree(csock);
   }
   sk->sk_protinfo = NULL;
}


//----------------------------------------------------------------------------//
// can_create()                                                               //
// create a CAN socket                                                        //
//----------------------------------------------------------------------------//
static int can_create(struct socket *sock, int protocol)
{
	struct sock * sk;
   
   PK_DBG(4, "sock=%p prot=%d", sock, protocol);

   
   //-----------------------------------------------------------------
   // create a new 'sock' structure and zero its contents (done by
   // setting the 3rd parameter to '1')
   //
   sk = sk_alloc(AF_CAN, GFP_ATOMIC, 1, NULL);
   if(sk == NULL)
   {
      return(-ENOMEM);
   }

      
   //-----------------------------------------------------------------
   // allocate new CAN sock structure
   //
   if(can_sock_alloc(sk) < 0)
   {
      kfree(sk);
      return(-ENOMEM);
   }

   //-----------------------------------------------------------------
   // init the socket
   //
   sock_init_data(sock, sk);
   sock->ops       = &can_raw_ops;
   sk->sk_protocol = protocol;
   // sk_set_owner(sk, THIS_MODULE);
   
   return 0;
}


//----------------------------------------------------------------------------//
// can_raw_accept()                                                           //
//                                                                            //
//----------------------------------------------------------------------------//
static int can_raw_accept(struct socket *sock, struct socket *newsock, 
                          int flags)
{
   PK_DBG(2, "socket %p", sock);
   return(0);
}


//----------------------------------------------------------------------------//
// can_raw_bind()                                                             //
//                                                                            //
//----------------------------------------------------------------------------//
static int can_raw_bind(struct socket *sock, struct sockaddr *myaddr, 
                        int sockaddr_len)
{
	PK_DBG(2, "socket %p", sock);
   return(0);
}


//----------------------------------------------------------------------------//
// can_raw_connect()                                                          //
// this functions is called upon "connect()" in userspace                     //
//----------------------------------------------------------------------------//
static int can_raw_connect(struct socket *sock, struct sockaddr *vaddr, 
                           int sockaddr_len, int flags)
{
   struct sockaddr_can *	ptsAddrCanT;
   struct can_sock *       ptsCanSockT;
   _TsCpHal *              ptsHalT;

      
   //-----------------------------------------------------------------
   // get pointer to can_sock structure 
   //
   ptsCanSockT = sock->sk->sk_protinfo;
   PK_DBG(eDBG_SOCK, "socket=%p flags=%d", sock, flags);
   PK_DBG(eDBG_SOCK, "can_sock=%p ", ptsCanSockT);
   
   //-----------------------------------------------------------------
   // test size of the supplied socket address structure
   //
   if(sockaddr_len > sizeof(struct sockaddr_can))
   {
      PK_DBG(eDBG_SOCK, "wrong size for sockaddr_can (%d)", sockaddr_len);
      return(-EADDRNOTAVAIL);    // Cannot assign requested address
   }
   
   
   //-----------------------------------------------------------------
   // vaddr points to a sockaddr_can structure
   //
   ptsAddrCanT = (struct sockaddr_can *) vaddr;

   PK_DBG(eDBG_SOCK, "CAN I/F = %d", ptsAddrCanT->can_if);
   PK_DBG(eDBG_SOCK, "CAN ID  = %X", ptsAddrCanT->can_id);

   
   //-----------------------------------------------------------------
   // check if there is a driver installed on the requested port
   //   
   ptsHalT = can_get_hal(ptsAddrCanT->can_if);
   if(ptsHalT == 0L)
   {
      PK_DBG(eDBG_SOCK, "No driver on CAN port %d", ptsAddrCanT->can_if);
      return(-EADDRNOTAVAIL);		// Cannot assign requested address
   }
   
   if( ( (ptsHalT->slStatus) & CP_HAL_DRV_OK) == 0)
   {
      PK_DBG(1, "No driver on CAN port %d", ptsAddrCanT->can_if);
      return(-EADDRNOTAVAIL);		// Cannot assign requested address
   }
   
   
   //-----------------------------------------------------------------
   // link the socket to the driver
   //
   ptsCanSockT->ptsCanPort = can_queue_create(  ptsAddrCanT->can_if, 
                                                ptsCanSockT->rcv_fifo_size, 
                                                ptsCanSockT->rcv_fifo_size);
   if(ptsCanSockT->ptsCanPort == 0L)
   {
      return(-ENOMEM);
   }
   
   return(0);
}


//----------------------------------------------------------------------------//
// can_raw_getname()                                                          //
//                                                                            //
//----------------------------------------------------------------------------//
static int can_raw_getname(struct socket *sock, struct sockaddr *addr, 
                           int *sockaddr_len, int peer)
{
   PK_DBG(eDBG_SOCK, "socket %p", sock);
   return(0);
}


//----------------------------------------------------------------------------//
// can_raw_getsockopt()                                                       //
//                                                                            //
//----------------------------------------------------------------------------//
static int can_raw_getsockopt(struct socket *sock, int level, int optname, 
                              char __user *optval, int __user *optlen)
{
   PK_DBG(eDBG_SOCK, "socket %p", sock);
   return(0);
}


//----------------------------------------------------------------------------//
// can_raw_ioctl()                                                            //
//                                                                            //
//----------------------------------------------------------------------------//
static int can_raw_ioctl(  struct socket *sock, unsigned int cmd, 
                           unsigned long arg)
{
   int                     slErrCodeT = 0;
   _TsCpBitrate            tsBitTimingT;
   uint32_t                aslValueT[8];        // buffer for ioctl values
   struct can_sock *       ptsCanSockT;
   
   //-----------------------------------------------------------------
   // get pointer to can_sock structure 
   //
   ptsCanSockT = sock->sk->sk_protinfo;

      
   switch(cmd)
   {
      //---------------------------------------------------
      // set new bitrate 
      //
      case SIOC_CAN_SET_BITRATE:
         PK_DBG(eDBG_SOCK, "Set bitrate");
         copy_from_user(&tsBitTimingT, (void __user *)arg, 
                        sizeof(tsBitTimingT));
         CpCoreBitrate(ptsCanSockT->ptsCanPort, &tsBitTimingT);
         
         break;
      
      //---------------------------------------------------
      // get current bitrate 
      //
      case SIOC_CAN_GET_BITRATE:
         PK_DBG(eDBG_SOCK, "Get bitrate");
   
         break;
   
      case SIOC_CAN_SET_MODE:
         PK_DBG(eDBG_SOCK, "Set CAN mode");
   
         break;
      
      case SIOC_CAN_GET_STATE:
         PK_DBG(eDBG_SOCK, "Get CAN state");
   
         break;

      //---------------------------------------------------
      // get max channel number (logical CAN I/F)
      //
      case SIOC_CAN_GET_CHANNEL_MAX:
         PK_DBG(eDBG_SOCK, "Get maximum CAN channel");
         aslValueT[0] = CP_CHANNEL_MAX;
         copy_to_user(  (void __user *)arg, 
                        &aslValueT[0],
                        sizeof(uint32_t));
         break;

      //---------------------------------------------------
      // set size of FIFO
      //
      case SIOC_CAN_SET_FIFOSIZE:
         PK_DBG(eDBG_SOCK, "Set Rcv-FIFO size");
         copy_from_user(   &aslValueT[0], 
                           (void __user *)arg, 
                           sizeof(uint32_t) * 2);
         //can_queue_resize(ptsCanSockT->ptsCanPort, aslValueT[0], aslValueT[1]);
         
         break;
      
   
      //---------------------------------------------------
      // set size of FIFO
      //
                  
      default:
         slErrCodeT = -EINVAL;
         break;
         
	}
   return(slErrCodeT);
}


//----------------------------------------------------------------------------//
// can_raw_listen()                                                           //
//                                                                            //
//----------------------------------------------------------------------------//
static int can_raw_listen(struct socket *sock, int len)
{
   PK_DBG(eDBG_SOCK, "socket %p", sock);
   return(0);
}


//----------------------------------------------------------------------------//
// can_raw_mmap()                                                             //
//                                                                            //
//----------------------------------------------------------------------------//
static int can_raw_mmap(struct file *file, struct socket *sock, 
                        struct vm_area_struct * vma)
{
   PK_DBG(eDBG_SOCK, "socket %p", sock);
   return(0);
}


//----------------------------------------------------------------------------//
// can_raw_poll()                                                             //
//                                                                            //
//----------------------------------------------------------------------------//
static unsigned int can_raw_poll(struct file *file, struct socket *sock, 
                                 struct poll_table_struct *wait)
{
   PK_DBG(eDBG_SOCK, "socket %p", sock);
   return(0);
}


//----------------------------------------------------------------------------//
// can_raw_recvmsg()                                                          //
//                                                                            //
//----------------------------------------------------------------------------//
static int can_raw_recvmsg(struct kiocb *iocb, struct socket *sock, 
                           struct msghdr *msg, size_t total_len, int flags)
{
   int                  slFrameCopyT;     // number of frames to copy
   int                  slFrameCntT;      // counter for copy loop
   int                  slFrameOffsetT;   // offset in the frame struct
   int                  slCpMsgCntT;      // messages copied from queue
   _U16                 uwMsgNumT;
   struct can_sock *    ptsCanSockT;
   _TsCpCanMsg          tsCanMsgT;        // CANpie frame 
   _TvCpStatus          tvCanStatusT;     // CAN status
      
   
   //----------------------------------------------------------------
   // Inside this function the CAN frames are copied from kernel
   // space into the user space. 
   // First, we calculate the total number of messages to copy,
   // i.e. the supplied buffer size divided by the '_TsCpCanMsg'
   // size.
   //
   slFrameCopyT = total_len / sizeof(_TsCpCanMsg);

   
   //-----------------------------------------------------------------
   // get pointer to can_sock structure 
   //
   ptsCanSockT = sock->sk->sk_protinfo;
   
   //----------------------------------------------------------------
   // set the counter for the queue to zero, it is incremented in
   // the loop
   //
   slCpMsgCntT = 0;

   
   //----------------------------------------------------------------
   // Try to copy the maximum number of possible messages, we quit
   // this loop if the CpCoreMsgRead() function does not return
   // one valid message
   //
   for(slFrameCntT = 0; slFrameCntT < slFrameCopyT; slFrameCntT++)
   {
      //---------------------------------------------------
      // try to get one message, if it fails, we quit the
      // loop
      //
      uwMsgNumT = 1; 
      tvCanStatusT = CpCoreMsgRead(ptsCanSockT->ptsCanPort, &tsCanMsgT, &uwMsgNumT);
      if(uwMsgNumT != 1) break;
      
      //---------------------------------------------------
      // increase the message counter
      //
      slCpMsgCntT++;
      PK_DBG(eDBG_DATA, "Message: ID=%04X DLC=%1d Data=%02X %02X", 
             tsCanMsgT.tuMsgId.ulExt, tsCanMsgT.ubMsgDLC, 
             tsCanMsgT.tuMsgData.aubByte[0],
             tsCanMsgT.tuMsgData.aubByte[1]);
      
      slFrameOffsetT = slFrameCntT * sizeof(_TsCpCanMsg);


      //---------------------------------------------------
      // copy the stuff to the user space
      //
      memcpy_toiovec(msg->msg_iov + slFrameOffsetT,
                     (char*) &tsCanMsgT, 
                     sizeof(_TsCpCanMsg));
      
   }
      
   
   //-----------------------------------------------------------------
   // return the length of the copied data
   //
   return(slCpMsgCntT * sizeof(_TsCpCanMsg));
}


//----------------------------------------------------------------------------//
// can_raw_release()                                                          //
//                                                                            //
//----------------------------------------------------------------------------//
static int can_raw_release(struct socket *sock)
{
	struct sock *     sk;
   struct can_sock * ptsCanSockT;

   sk = sock->sk;

      
   //-----------------------------------------------------------------
   // get pointer to can_sock structure 
   //
   ptsCanSockT = sock->sk->sk_protinfo;
   PK_DBG(eDBG_SOCK, "can_sock=%p ", ptsCanSockT);

   can_queue_delete(ptsCanSockT->ptsCanPort);
   
   PK_DBG(eDBG_SOCK, "sock=%p", sk);
	can_sock_release(sk);
   //ptsCanSockT = (struct can_sock_raw*) ptsSocketV->sk;
   //kfree(ptsCanSockT);

   return(0);
}


//----------------------------------------------------------------------------//
// can_raw_sendmsg()                                                          //
//                                                                            //
//----------------------------------------------------------------------------//
static int can_raw_sendmsg(struct kiocb *iocb, struct socket *sock, 
                           struct msghdr *msg, size_t total_len)
{
   int                  slFrameNumT;      // number of frames to copy
   //int                  slFrameCntT;      // counter for copy loop
   //int                  slMsgOffsetT;
   _U16                 uwMsgCntT;
   struct can_sock *    ptsCanSockT;
   //_TsCpCanMsg          tsCanMsgT;        // CANpie frame 
   char *               pszBufferT;
   _TsCpCanMsg *        ptsFrameT;        // CANpie frame 
      
   
   PK_DBG(eDBG_SOCK, "socket %p", sock);
   
   //---------------------------------------------------------------
   // data is transfered via the structure msghdr (linux/socket.h)
   //
   if(msg->msg_flags & ~MSG_DONTWAIT)
   {
      return(-EINVAL);
   }
   /*
   if (sk->sk_state != TCP_ESTABLISHED)
      return -ENOTCONN;
   */

   //----------------------------------------------------------------
   // calculate the number of messages to copy
   //
   slFrameNumT = total_len / sizeof(_TsCpCanMsg);
   PK_DBG(eDBG_DATA, "copy %d bytes", total_len);
   
   pszBufferT = kmalloc(total_len, GFP_ATOMIC);
   
   //-----------------------------------------------------------------
   // get pointer to can_sock structure 
   //
   ptsCanSockT = sock->sk->sk_protinfo;
   
   //----------------------------------------------------------------
   // copy data into CANpie message structure and call the 
   // CANpie alculate the number of messages to copy
   //
   PK_DBG(eDBG_DATA, "copy %d CAN frames", slFrameNumT);
   /*
   for(slFrameCntT = 0; slFrameCntT < slFrameNumT; slFrameCntT++)
   {
      slMsgOffsetT = slFrameCntT * sizeof(_TsCpCanMsg);
      PK_DBG(eDBG_DATA, "offset = %d", slMsgOffsetT);
      memcpy_fromiovec( (char*) &tsCanMsgT, 
                        msg->msg_iov ,//+ slMsgOffsetT, 
                        sizeof(_TsCpCanMsg));
       msg->msg_iov =  msg->msg_iov + sizeof(_TsCpCanMsg);
      uwMsgCntT = 1;
      
      PK_DBG(eDBG_DATA, "Message: ID=%04X DLC=%1d Data=%02X %02X", 
             tsCanMsgT.tuMsgId.ulExt, tsCanMsgT.ubMsgDLC, 
             tsCanMsgT.tuMsgData.aubByte[0],
             tsCanMsgT.tuMsgData.aubByte[1]);
      CpCoreMsgWrite(ptsCanSockT->ptsCanPort, &tsCanMsgT, &uwMsgCntT);
   }
    */ 
   memcpy_fromiovec( pszBufferT, msg->msg_iov , total_len);
   ptsFrameT = (_TsCpCanMsg *) pszBufferT;
   uwMsgCntT = slFrameNumT;
   PK_DBG(eDBG_DATA, "Message: ID=%04X DLC=%1d Data=%02X %02X", 
             ptsFrameT->tuMsgId.ulExt, ptsFrameT->ubMsgDLC, 
             ptsFrameT->tuMsgData.aubByte[0],
             ptsFrameT->tuMsgData.aubByte[1]);
   
   CpCoreMsgWrite(ptsCanSockT->ptsCanPort, ptsFrameT, &uwMsgCntT);
   kfree(pszBufferT);
   //-----------------------------------------------------------------
   // return the length of the copied data
   //
   return(total_len);
}


//----------------------------------------------------------------------------//
// can_raw_sendpage()                                                         //
//                                                                            //
//----------------------------------------------------------------------------//
static ssize_t can_raw_sendpage(struct socket *sock, struct page *page, 
                                int offset, size_t size, int flags)
{
   PK_DBG(eDBG_SOCK, "socket %p", sock);
   return(0);
}


//----------------------------------------------------------------------------//
// can_raw_setsockopt()                                                       //
//                                                                            //
//----------------------------------------------------------------------------//
static int can_raw_setsockopt(struct socket *sock, int level, 
                              int optname, char __user *optval, int optlen)
{
   PK_DBG(eDBG_SOCK, "Level=%d  Option=%d", level, optname);
   return(0);
}


//----------------------------------------------------------------------------//
// can_raw_shutdown()                                                         //
//                                                                            //
//----------------------------------------------------------------------------//
static int can_raw_shutdown(struct socket *sock, int flags)
{
   PK_DBG(eDBG_SOCK, "socket %p", sock);
   return(0);
}


//----------------------------------------------------------------------------//
// can_raw_socketpair()                                                       //
//                                                                            //
//----------------------------------------------------------------------------//
static int can_raw_socketpair(struct socket *sock1, struct socket *sock2)
{
   return(0);
}



/*----------------------------------------------------------------------------*\
** Static structures                                                          **
** proto_ops / net_proto_family                                               **
\*----------------------------------------------------------------------------*/

static struct proto_ops can_raw_ops = {
      
   .family 		= AF_CAN,
   .owner      = THIS_MODULE,
   .release    = can_raw_release,
   .bind       = can_raw_bind,
   .connect    = can_raw_connect,
   .socketpair = can_raw_socketpair,
   .accept     = can_raw_accept,
   .getname    = can_raw_getname,
   .poll       = can_raw_poll,
   .ioctl      = can_raw_ioctl,
   .listen     = can_raw_listen,
   .shutdown   = can_raw_shutdown,
   .setsockopt = can_raw_setsockopt,
   .getsockopt = can_raw_getsockopt,
   .sendmsg    = can_raw_sendmsg,
   .recvmsg    = can_raw_recvmsg,
   .mmap       = can_raw_mmap,
   .sendpage   = can_raw_sendpage,
};

static struct net_proto_family can_family_ops =
{
	AF_CAN,
	can_create
};


//----------------------------------------------------------------------------//
// can_socket_init()                                                          //
// init CAN protocol layer                                                    //
//----------------------------------------------------------------------------//
int can_socket_init(void)
{
   
   sock_register(&can_family_ops);

	return 0;
   
}


//----------------------------------------------------------------------------//
// can_socket_cleanup()                                                       //
// remove CAN protocol layer                                                  //
//----------------------------------------------------------------------------//
void can_socket_cleanup(void)
{

   sock_unregister(AF_CAN);

}


