//****************************************************************************//
// File:          canpie_hal.c                                                //
// Description:   CANpie hardware abstraction layer (HAL)                     //
// Author:        Uwe Koppe                                                   //
// e-mail:        koppe@microcontrol.net                                      //
//                                                                            //
//============================================================================//
// This program is free software; you can redistribute it and/or modify       //
// it under the terms of the GNU Lesser General Public License as published   //
// by the Free Software Foundation; either version 2.1 of the License, or     //
// (at your option) any later version.                                        //
//============================================================================//
//                                                                            //
// Date        History                                                        //
// ----------  -------------------------------------------------------------- //
// 29.07.2003  Initial version                                                //
//                                                                            //
//****************************************************************************//


//------------------------------------------------------------------------------
// CVS version information:
// $Id: canpie_hal.c,v 1.4 2005/06/23 13:29:54 microcontrol Exp $
//------------------------------------------------------------------------------




/*----------------------------------------------------------------------------*\
** Include files                                                              **
**                                                                            **
\*----------------------------------------------------------------------------*/

#include <asm/uaccess.h>         // memset()

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>

#include <net/can/canpie_dbg.h>
#include <net/can/canpie_hal.h>           // Hardware abstraction layer
#include <net/can/canpie_proc.h>          // interface to /proc filesystem
#include <net/can/cpcanmsg.h>             // CAN message struct access

#if CP_CONFIG_SOCKET == 1
#include <net/can/af_can.h>               // socket support
#endif


/*----------------------------------------------------------------------------*\
** Definitions                                                                **
**                                                                            **
\*----------------------------------------------------------------------------*/




/*----------------------------------------------------------------------------*\
** Module information                                                         **
**                                                                            **
\*----------------------------------------------------------------------------*/
MODULE_DESCRIPTION("CANpie kernel driver");
MODULE_AUTHOR("Uwe Koppe <koppe@microcontrol.net>");
MODULE_LICENSE("GPL");
MODULE_VERSION("1:"CP_MODULE_VERSION);


EXPORT_SYMBOL(can_register_driver);
EXPORT_SYMBOL(can_unregister_driver);

EXPORT_SYMBOL(can_queue_rcv_pop);
EXPORT_SYMBOL(can_queue_rcv_push);
EXPORT_SYMBOL(can_queue_trm_pop);
EXPORT_SYMBOL(can_queue_trm_push);

/*----------------------------------------------------------------------------*\
** Module variables                                                           **
**                                                                            **
\*----------------------------------------------------------------------------*/

static _TsCpHal   atsCpHalS[CP_CHANNEL_MAX];


/*----------------------------------------------------------------------------*\
** Debugging                                                                  **
**                                                                            **
\*----------------------------------------------------------------------------*/
#ifdef  CP_DEBUG
unsigned short    uwCpDebugG = 2;
module_param_named(debug, uwCpDebugG, ushort, 0);
MODULE_PARM_DESC(debug, " Set debug level, 0=quiet .. 5=noisy");

static char szMsgDbgS[64];    // CAN message debugging -> can_debug_frame()
static char *aszBitratesS[] = {  "10 kBit/s",
                                 "20 kBit/s",
                                 "50 kBit/s",
                                 "100 kBit/s",
                                 "125 kBit/s",
                                 "250 kBit/s",
                                 "500 kBit/s",
                                 "800 kBit/s",
                                 "1 MBit/s",
                                 "Used defined, BTR values"
                              };
                                 
static char *aszModeS[] =     {  "STOP",
                                 "ACTIVE",
                                 "SLEEP"
                              };
#endif


#ifdef  CP_DEBUG
static void can_print_frame(_TsCpCanMsg * ptsCanMsgV, char * szBufferV);
#else
#define can_print_frame(A, B)    
#endif



/*----------------------------------------------------------------------------*\
** Functions                                                                  **
**                                                                            **
\*----------------------------------------------------------------------------*/



//----------------------------------------------------------------------------//
// CpCoreBitrate()                                                            //
// Setup baudrate of CAN controller                                           //
//----------------------------------------------------------------------------//
_TvCpStatus CpCoreBitrate(_TsCpPort * ptsPortV, _TsCpBitrate * ptsBitrateV)
{
   _U08           ubReturnT;
   _FnBitrate     pfnHalT;    // function of HAL
   

   //----------------------------------------------------------------
   // check logic port number
   //
   if(ptsPortV == 0L) return(CpErr_CHANNEL);   
   if( (ptsPortV->slLogIf) >= CP_CHANNEL_MAX) return (CpErr_CHANNEL);
   
   //----------------------------------------------------------------
   // test if handler is installed on the requested channel
   //
   pfnHalT = atsCpHalS[ptsPortV->slLogIf].pfnBitrate;
   if(pfnHalT == 0L)
   {
      PK_DBG(eDBG_FUNC, "can%d: No driver function", ptsPortV->slLogIf);
      return(CpErr_NOT_SUPPORTED);
   }

   if(ptsBitrateV->ubBaudSel > CP_BAUD_MAX)
   {
      PK_DBG(  eDBG_FUNC, "can%d: bitrate out of range (ubBaudSel=%d)", 
               ptsPortV->slLogIf, ptsBitrateV->ubBaudSel                );
      return(CpErr_BAUDRATE); 
   }
   
   
   //----------------------------------------------------------------
   // call the handler 
   //
   //    
   PK_DBG(  eDBG_FUNC, "can%d: set bitrate to %s", 
            ptsPortV->slLogIf, aszBitratesS[ptsBitrateV->ubBaudSel]     );
            
   ubReturnT = (* pfnHalT )(ptsPortV, ptsBitrateV);         
   return(ubReturnT);
}


//----------------------------------------------------------------------------//
// CpCoreBufferGetDlc()                                                   		//
// Get DLC of buffer                                                			   //
//----------------------------------------------------------------------------//
_TvCpStatus CpCoreBufferGetDlc(_TsCpPort * ptsPortV, _U08 ubBufferV, _U08 * pubDlcV)
{
   _U08              ubReturnT;
   _FnBufferGetDlc   pfnHalT;          // function of HAL
   

   //----------------------------------------------------------------
   // check logic port number
   //
   if(ptsPortV == 0L) return(CpErr_CHANNEL);   
   if( (ptsPortV->slLogIf) >= CP_CHANNEL_MAX) return (CpErr_CHANNEL);
   
   //----------------------------------------------------------------
   // test if handler is installed on the requested channel
   //
   pfnHalT = atsCpHalS[ptsPortV->slLogIf].pfnBufferGetDlc;
   if(pfnHalT == 0L)
   {
      PK_DBG(2, "NULL pointer on channel %d", ptsPortV->slLogIf);
      return(CpErr_NOT_SUPPORTED);
   }


   //----------------------------------------------------------------
   // call the handler 
   //
   ubReturnT = (* pfnHalT )(ptsPortV, ubBufferV, pubDlcV);         
   return(ubReturnT);
}


//----------------------------------------------------------------------------//
// CpCoreBufferInit()                                                        //
// Allocate buffer of FullCAN controller                                      //
//----------------------------------------------------------------------------//
_TvCpStatus CpCoreBufferInit(_U08 ubChannelV, _TsCpCanMsg * ptsCanMsgV,
                      _U08 ubDirectionV)
{
   //---	test the channel number ------------------------------------
	if( ubChannelV >= CP_CHANNEL_MAX) return (CpErr_CHANNEL);


   return (CpErr_OK);
}

//----------------------------------------------------------------------------//
// CpCoreBufferRelease()                                                      //
// De-Allocate a FullCAN buffer                                               //
//----------------------------------------------------------------------------//
_TvCpStatus CpCoreBufferRelease(_U08 ubChannelV, _U08 ubBufferNumV)
{

   //---	test the channel number ------------------------------------
	if( (ubChannelV + 1) > CP_CHANNEL_MAX) return (CpErr_CHANNEL);

	return (CpErr_OK);
}


//----------------------------------------------------------------------------//
// CpCoreBufferSetDlc()                                                   		//
// Set DLC of buffer                                                			   //
//----------------------------------------------------------------------------//
_TvCpStatus CpCoreBufferSetDlc(	_U08 ubChannelV, _U08 ubBufferV, _U08 ubDlcV)
{
   //--- test the channel number ------------------------------------
   if( ubChannelV >= CP_CHANNEL_MAX) return (CpErr_CHANNEL);



   return (CpErr_OK);
}




//----------------------------------------------------------------------------//
// CpCoreBufferSend()                                                     		//
// Send message of FullCAN buffer                                      			//
//----------------------------------------------------------------------------//
_TvCpStatus CpCoreBufferSend(_U08 ubChannelV, _U08 ubBufferV)
{
   //--- test the channel number ------------------------------------
   if( ubChannelV >= CP_CHANNEL_MAX) return (CpErr_CHANNEL);

   return (CpErr_OK);
}


//----------------------------------------------------------------------------//
// CpCoreBufferTransmit()                                                     //
// Transmit contents of buffer                                                //

//----------------------------------------------------------------------------//
_TvCpStatus CpCoreBufferTransmit(_U08 ubChannelV, _TsCpCanMsg * ptsCanMsgV)
{
   //--- test the channel number ------------------------------------
   if( (ubChannelV + 1) > CP_CHANNEL_MAX) return (CpErr_CHANNEL);

	return (CpErr_OK);
}


//----------------------------------------------------------------------------//
// CpCoreCanMode()                                                            //
// Set state of CAN controller                                                //
//----------------------------------------------------------------------------//
_TvCpStatus CpCoreCanMode(_TsCpPort * ptsPortV, _U08 ubModeV)
{
   _U08              ubReturnT;
   _FnCanMode        pfnHalT;          // function of HAL
   

   //----------------------------------------------------------------
   // check logic port number
   //
   if(ptsPortV == 0L) return(CpErr_CHANNEL);   
   if( (ptsPortV->slLogIf) >= CP_CHANNEL_MAX) return (CpErr_CHANNEL);
   
   //----------------------------------------------------------------
   // test if handler is installed on the requested channel
   //
   pfnHalT = atsCpHalS[ptsPortV->slLogIf].pfnCanMode;
   if(pfnHalT == 0L)
   {
      PK_DBG(eDBG_FUNC, "can%d: No driver function", ptsPortV->slLogIf);
      return(CpErr_NOT_SUPPORTED);
   }

   if(ubModeV > CP_MODE_SLEEP)
   {
      PK_DBG(  eDBG_FUNC, "can%d: mode out of range (mode=%d)", 
               ptsPortV->slLogIf, ubModeV                         );
      return(CpErr_PARAM); 
   }
   

   //----------------------------------------------------------------
   // call the handler 
   //
   PK_DBG(  eDBG_FUNC, "can%d: set mode to '%s'", 
            ptsPortV->slLogIf, aszModeS[ubModeV]   );
   ubReturnT = (* pfnHalT )(ptsPortV, ubModeV);         
   return(ubReturnT);
}


//----------------------------------------------------------------------------//
// CpCoreCanState()                                                           //
// Retrive state of CAN controller                                            //
//----------------------------------------------------------------------------//
_TvCpStatus CpCoreCanStatus(_TsCpPort * ptsPortV, _U08 * pubStatusV)
{
   _U08              ubReturnT;
   _FnCanStatus      pfnHalT;          // function of HAL
   

   //----------------------------------------------------------------
   // check logic port number
   //
   if(ptsPortV == 0L) return(CpErr_CHANNEL);   
   if( (ptsPortV->slLogIf) >= CP_CHANNEL_MAX) return (CpErr_CHANNEL);
   
   //----------------------------------------------------------------
   // test if handler is installed on the requested channel
   //
   pfnHalT = atsCpHalS[ptsPortV->slLogIf].pfnCanStatus;
   if(pfnHalT == 0L)
   {
      PK_DBG(2, "NULL pointer on channel %d", ptsPortV->slLogIf);
      return(CpErr_NOT_SUPPORTED);
   }


   //----------------------------------------------------------------
   // call the handler 
   //
   ubReturnT = (* pfnHalT )(ptsPortV, pubStatusV);
   return(ubReturnT);
}


//----------------------------------------------------------------------------//
// CpCoreCanStatistic()                                                       //
//                                                                            //
//----------------------------------------------------------------------------//
_TvCpStatus CpCoreCanStatistic(_TsCpPort * ptsPortV, _TsCpStats * ptsStatV)
{
   _U08              ubReturnT;
   _FnCanStats       pfnHalT;          // function of HAL
   

   //----------------------------------------------------------------
   // check logic port number
   //
   if(ptsPortV == 0L) return(CpErr_CHANNEL);   
   if( (ptsPortV->slLogIf) >= CP_CHANNEL_MAX) return (CpErr_CHANNEL);
   
   //----------------------------------------------------------------
   // test if handler is installed on the requested channel
   //
   pfnHalT = atsCpHalS[ptsPortV->slLogIf].pfnCanStats;
   if(pfnHalT == 0L)
   {
      PK_DBG(2, "NULL pointer on channel %d", ptsPortV->slLogIf);
      return(CpErr_NOT_SUPPORTED);
   }


   //----------------------------------------------------------------
   // call the handler 
   //
   ubReturnT = (* pfnHalT )(ptsPortV, ptsStatV);         
   return(ubReturnT);
}


//----------------------------------------------------------------------------//
// CpCoreDriverInit()                                                         //
// Initialize the CAN driver                                                  //
//----------------------------------------------------------------------------//
_TvCpStatus CpCoreDriverInit(_U08 ubLogIfV, _TsCpPort * ptsPortV)
{
   _U08           ubReturnT;
   _U08           ubPhyIfT;   // physical CAN interface (for driver)
   _FnDriverInit  pfnHalT;    // function of HAL
   
   
   //----------------------------------------------------------------
   // test the logical CAN interface number
   //
   if( ubLogIfV >= CP_CHANNEL_MAX) return (CpErr_CHANNEL);


   //----------------------------------------------------------------
   // test if handler is installed on the requested channel
   //
   pfnHalT = atsCpHalS[ubLogIfV].pfnDriverInit;
   if(pfnHalT == 0L)
   {
      PK_DBG(eDBG_FUNC, "No driver function on can%d", ubLogIfV);
      return(CpErr_NOT_SUPPORTED);
   }


   //----------------------------------------------------------------
   // convert the logical interface number to the physical interface
   // number and call the function of the driver
   //
   ubPhyIfT = ptsPortV->slPhyIf;
   ubReturnT = (* pfnHalT )(ubPhyIfT, ptsPortV);

   return(ubReturnT);
}


//----------------------------------------------------------------------------//
// CpCoreDriverRelease()                                                      //
// release the CAN driver                                                     //
//----------------------------------------------------------------------------//
_TvCpStatus CpCoreDriverRelease(_TsCpPort * ptsPortV)
{
   _U08              ubReturnT;
   _FnDriverRelease  pfnHalT;          // function of HAL
   

   //----------------------------------------------------------------
   // check logic port number
   //
   if(ptsPortV == 0L) return(CpErr_CHANNEL);   
   if( (ptsPortV->slLogIf) >= CP_CHANNEL_MAX) return (CpErr_CHANNEL);
   
   //----------------------------------------------------------------
   // test if handler is installed on the requested channel
   //
   pfnHalT = atsCpHalS[ptsPortV->slLogIf].pfnDriverRelease;
   if(pfnHalT == 0L)
   {
      PK_DBG(2, "NULL pointer on channel %d", ptsPortV->slLogIf);
      return(CpErr_NOT_SUPPORTED);
   }


   //----------------------------------------------------------------
   // call the handler 
   //
   ubReturnT = (* pfnHalT )(ptsPortV);         
   return(ubReturnT);
}



//----------------------------------------------------------------------------//
// CpCoreHDI()                                                                //
// Fill Hardware Description Interface (HDI) structure                        //
//----------------------------------------------------------------------------//
_TvCpStatus CpCoreHDI(_U08 ubChannelV, CpStruct_HDI * pHdiV)
{
   //--- test the channel number ------------------------------------

   if( (ubChannelV + 1) > CP_CHANNEL_MAX) return (CpErr_CHANNEL);

   return (CpErr_OK);
}


//----------------------------------------------------------------------------//
// CpCoreFilterAll()                                                          //
// Enable / Disable filter for all messages                                   //
//----------------------------------------------------------------------------//
_TvCpStatus CpCoreFilterAll(_U08 ubChannelV, _BIT btEnableV)
{
   //--- test the channel number ------------------------------------
   if( (ubChannelV + 1) > CP_CHANNEL_MAX) return (CpErr_CHANNEL);


   return (CpErr_OK);
}


//----------------------------------------------------------------------------//
// CpCoreFilterMsg()                                                          //
// Enable / Disable filter for a single messages                              //
//----------------------------------------------------------------------------//
_TvCpStatus CpCoreFilterMsg(_U08 ubChannelV, _U16 uwIdV, _BIT btEnableV)
{
   //--- test the channel number ------------------------------------
   if( ubChannelV >= CP_CHANNEL_MAX) return (CpErr_CHANNEL);


   return (CpErr_OK);
}


//----------------------------------------------------------------------------//
// CpCoreMsgReceive()                                                         //
// Read CAN message from controller                                           //
//----------------------------------------------------------------------------//
_TvCpStatus CpCoreMsgRead( _TsCpPort * ptsPortV, _TsCpCanMsg * ptsCanMsgV, 
                           _U16 * puwMsgCntV)
{
   _U08              ubReturnT;
   _FnMsgRead        pfnHalT;          // function of HAL
   

   //----------------------------------------------------------------
   // check logic port number
   //
   if(ptsPortV == 0L) return(CpErr_CHANNEL);   
   if( (ptsPortV->slLogIf) >= CP_CHANNEL_MAX) return (CpErr_CHANNEL);
   
   //----------------------------------------------------------------
   // test if handler is installed on the requested channel
   //
   pfnHalT = atsCpHalS[ptsPortV->slLogIf].pfnMsgRead;
   if(pfnHalT == 0L)
   {
      PK_DBG(eDBG_FUNC, "No driver function on can%d", ptsPortV->slLogIf);
      return(CpErr_NOT_SUPPORTED);
   }


   //----------------------------------------------------------------
   // call the handler 
   //
   ubReturnT = (* pfnHalT )(ptsPortV, ptsCanMsgV, puwMsgCntV);         
   return(ubReturnT);
}


//----------------------------------------------------------------------------//
// CpCoreMsgWrite()                                                           //
// Write CAN message to controller                                            //
//----------------------------------------------------------------------------//
_TvCpStatus CpCoreMsgWrite(_TsCpPort * ptsPortV, _TsCpCanMsg * ptsCanMsgV,
                           _U16 * puwMsgCntV)
{
   _U08              ubReturnT;
   int               slMsgCntReqT;     // number of messages to send
   int               slFifoCopyT;      // number of messages copied to FIFO
   _FnMsgWrite       pfnHalT;          // function of HAL
   

   //----------------------------------------------------------------
   // check logic port number
   //
   if(ptsPortV == 0L) return(CpErr_CHANNEL);   
   if( (ptsPortV->slLogIf) >= CP_CHANNEL_MAX) return (CpErr_CHANNEL);
   
   //----------------------------------------------------------------
   // test if handler is installed on the requested channel
   //
   pfnHalT = atsCpHalS[ptsPortV->slLogIf].pfnMsgWrite;
   if(pfnHalT == 0L)
   {
      PK_DBG(eDBG_FUNC, "No driver function on can%d", ptsPortV->slLogIf);
      return(CpErr_NOT_SUPPORTED);
   }

   //----------------------------------------------------------------
   // push the messages into the transmit queue, a positiv
   // return value denotes the number of messages that could
   // be placed into the transmit FIFO. In general, the value
   // should be the same then the request. A negative return
   // value denotes an error.
   //
   slMsgCntReqT = (int) (*puwMsgCntV);
   if(slMsgCntReqT < 1)
   {
      //---------------------------------------------------
      // there is nothing to send, so just quit here
      //
      return(CpErr_OK);   
   }
   slFifoCopyT = can_queue_trm_push(ptsPortV, ptsCanMsgV, slMsgCntReqT);

   can_print_frame(ptsCanMsgV, szMsgDbgS);
   PK_DBG(eDBG_FUNC, "%s", szMsgDbgS);
   
   //----------------------------------------------------------------
   // test for error return codes
   //
   if(slFifoCopyT < 0)
   {
      if(slFifoCopyT == -ENODEV) return(CpErr_CHANNEL);
      if(slFifoCopyT == -EINVAL) return(CpErr_FIFO_PARM);
      
      return(CpErr_GENERIC);
   }

      
   //----------------------------------------------------------------
   // test if requested size is equal to response size
   //
   if(slFifoCopyT != slMsgCntReqT)
   {
      ubReturnT = CpErr_FIFO_FULL;
   }
   else
   {
      ubReturnT = CpErr_OK;
   }
   
   
   //----------------------------------------------------------------
   // call the handler 
   //
   (* pfnHalT )(ptsPortV, ptsCanMsgV, puwMsgCntV);         
   return(ubReturnT);
}



//----------------------------------------------------------------------------//
// CpCoreRegRead()                                                            //
// Read one byte from specified CAN controller register                       //
//----------------------------------------------------------------------------//
_TvCpStatus CpCoreRegRead(_U08 ubChannelV, _U16 uwRegAdrV, _U08 * pubValueV)
{
   //--- check if channel is valid ----------------------------------
   if (ubChannelV >= CP_CHANNEL_MAX) return (CpErr_CHANNEL);


   return (CpErr_OK);
}


//----------------------------------------------------------------------------//
// CpCoreRegWrite()                                                           //
// Write one byte to specified CAN controller register                        //
//----------------------------------------------------------------------------//
_TvCpStatus CpCoreRegWrite(_U08 ubChannelV, _U16 uwRegAdrV, _U08 ubValueV)
{
   //--- check if channel is valid ----------------------------------
   if (ubChannelV >= CP_CHANNEL_MAX) return (CpErr_CHANNEL);


   return (CpErr_OK);
}



//============================================================================//
// functions for queue management                                             //
//                                                                            //
//============================================================================//


//----------------------------------------------------------------------------//
// can_queue_create()                                                         //
// create a new Tx & Rx queue on logic CAN interface with index 'slLogIfV'    //
//----------------------------------------------------------------------------//
_TsCpPort * can_queue_create(int slLogIfV, 
                             int slRcvFifoSizeV, int slTrmFifoSizeV)
{
   _TsCpMsgQueue *   ptsQueueT;           // pointer to CAN message queue
   int               slQueueCntT;         // counter for queue
   int               slQueueIdT = -1;     // selected queue ID
   
      
   
   PK_DBG(eDBG_PORT, "Create queue on log. CAN-I/F %d", slLogIfV);

   //----------------------------------------------------------------
   // check if the logical interface number is in range
   //   
   if( (slLogIfV > CP_CHANNEL_MAX - 1) || (slLogIfV < 0) )
   {
      return(0L);
   }

         
   //-------------------------------------------------------------------
   // check the FIFO size first and make sure that it is not above
   // 65535 entries
   //
   if( (slRcvFifoSizeV < CP_QUEUE_MSG_MIN) || 
       (slRcvFifoSizeV > CP_QUEUE_MSG_MAX)   )
   {
      slRcvFifoSizeV = CP_QUEUE_MSG_DEF;
   }

   if( (slTrmFifoSizeV < CP_QUEUE_MSG_MIN) || 
       (slTrmFifoSizeV > CP_QUEUE_MSG_MAX)   )
   {
      slTrmFifoSizeV = CP_QUEUE_MSG_DEF;
   }
   
   //----------------------------------------------------------------
   // check if there is a free entry in the queue array for the
   // selected logic CAN interface
   //
   for(slQueueCntT = 0; slQueueCntT < CP_QUEUE_MAX; slQueueCntT++)
   {
      ptsQueueT = &(atsCpHalS[slLogIfV].atsMsgQueue[slQueueCntT]);
      PK_DBG(2, "slQueue = %d", ptsQueueT->tsCanPort.slQueue);
      
      if(ptsQueueT->tsCanPort.slQueue == -1)
      {
         //------------------------------------------------
         // this queue is not assigned, store the queue
         // number and quit the loop
         //
         slQueueIdT = slQueueCntT;
         PK_DBG(2, "found free queue, ID= %d", slQueueIdT);
         break;
      }
   }

   //----------------------------------------------------------------
   // check if a free queue was found
   //
   if(slQueueIdT < 0) 
   {
      PK_DBG(2, "no free queue found");
      return(0L);
   }

   //----------------------------------------------------------------
   // set pointer to the selected message queue
   //
   ptsQueueT = &(atsCpHalS[slLogIfV].atsMsgQueue[slQueueIdT]);
   
   //----------------------------------------------------------------
   // create the message queue, i.e. init FIFOs for receive and
   // transmit
   // 
   if( CpFifoInit( &(ptsQueueT->tsRcvFifo), (_U16) slRcvFifoSizeV) != CpErr_OK)
   {
      goto QueueFail;
   }
      
   if( CpFifoInit( &(ptsQueueT->tsTrmFifo), (_U16) slTrmFifoSizeV) != CpErr_OK)
   {
      CpFifoRelease(&(ptsQueueT->tsRcvFifo));
      goto QueueFail;
   }

   //----------------------------------------------------------------
   // mark the message queue as assigned
   //
   ptsQueueT->tsCanPort.slLogIf = slLogIfV;
   ptsQueueT->tsCanPort.slQueue = slQueueIdT;

   atsCpHalS[slLogIfV].slUseCount++;
      
   return( &(ptsQueueT->tsCanPort) );
   
QueueFail:
   
   return(0L);
}


//----------------------------------------------------------------------------//
// can_queue_delete()                                                         //
//                                                                            //
//----------------------------------------------------------------------------//
int can_queue_delete(_TsCpPort * ptsCanPortV)
{
   int               slLogIfT;            // number of logical CAN interface   
   int               slQueueIdT = -1;     // selected queue ID
   _TsCpMsgQueue *   ptsQueueT;           // pointer to CAN message queue
   
   
   //----------------------------------------------------------------
   // test the _TsCpPort pointer
   //
   if(ptsCanPortV == 0L)
   {
      return(-EINVAL);  
   }
   
   
   //----------------------------------------------------------------
   // test the logical interface number
   //
   slLogIfT = ptsCanPortV->slLogIf;
   if( (slLogIfT > CP_CHANNEL_MAX - 1) || (slLogIfT < 0) )
   {
      PK_DBG(2, "No CAN interface on port %d", slLogIfT);
      return(-ENODEV);
   }

   //----------------------------------------------------------------
   // test the queue ID number
   //
   slQueueIdT = ptsCanPortV->slQueue;
   if( (slQueueIdT < 0) || (slQueueIdT > CP_QUEUE_MAX - 1) )
   {
      PK_DBG(2, "No queue ID %d on CAN interface %d", slQueueIdT, slLogIfT);
      return(-EINVAL);  
   }
   
   //----------------------------------------------------------------
   // select the message queue and delete the FIFOs
   //
   ptsQueueT = &(atsCpHalS[slLogIfT].atsMsgQueue[slQueueIdT]);
   CpFifoRelease( &(ptsQueueT->tsRcvFifo) );
   CpFifoRelease( &(ptsQueueT->tsTrmFifo) );
   
   //----------------------------------------------------------------
   // mark the queue as not assigned
   //
   ptsCanPortV->slQueue = -1;
   atsCpHalS[slLogIfT].slUseCount--;
   return(0);
}


//----------------------------------------------------------------------------//
// can_queue_rcv_pop()                                                        //
// read CAN messages from specified CAN port (queue)                          //
//----------------------------------------------------------------------------//
int can_queue_rcv_pop(  _TsCpPort * ptsCanPortV, 
                        _TsCpCanMsg * ptsCanMsgV, int slMsgNumV)
{
   int               slLogIfT;            // number of logical CAN interface   
   int               slQueueIdT = -1;     // selected queue ID
   int               slMsgCopyT;          // number of messages copied
   _U08              ubFifoStatusT;       // status of receive FIFO      
   _TsCpMsgQueue *   ptsQueueT;           // pointer to CAN message queue
   
   
   //----------------------------------------------------------------
   // test the _TsCpPort pointer
   //
   if(ptsCanPortV == 0L)
   {
      return(-EINVAL);  
   }
   
   
   //----------------------------------------------------------------
   // test the logical interface number
   //
   slLogIfT = ptsCanPortV->slLogIf;
   if( (slLogIfT > CP_CHANNEL_MAX - 1) || (slLogIfT < 0) )
   {
      PK_DBG(2, "No CAN interface on port %d", slLogIfT);
      return(-ENODEV);
   }

   //----------------------------------------------------------------
   // test the queue ID number
   //
   slQueueIdT = ptsCanPortV->slQueue;
   if( (slQueueIdT < 0) || (slQueueIdT > CP_QUEUE_MAX - 1) )
   {
      PK_DBG(2, "No queue ID %d on CAN interface %d", slQueueIdT, slLogIfT);
      return(-EINVAL);  
   }

   //----------------------------------------------------------------
   // select the message queue and read 'slMsgNumV' messages
   //
   ptsQueueT = &(atsCpHalS[slLogIfT].atsMsgQueue[slQueueIdT]);
   for(slMsgCopyT = 0; slMsgCopyT < slMsgNumV; slMsgCopyT++)
   {
      ubFifoStatusT = CpFifoPop(&(ptsQueueT->tsRcvFifo), ptsCanMsgV);
      if(ubFifoStatusT == CpErr_FIFO_EMPTY) break;
      ptsCanMsgV++;
   }
   
   return(slMsgCopyT);  
}


//----------------------------------------------------------------------------//
// can_queue_rcv_push()                                                       //
// write CAN messages to specified CAN port ( all open queues)                //
//----------------------------------------------------------------------------//
int can_queue_rcv_push( _TsCpPort * ptsCanPortV, 
                        _TsCpCanMsg * ptsCanMsgV, int slMsgNumV)
{
   int               slLogIfT;            // number of logical CAN interface   
   int               slQueueCntT;         // counter for message queue
   int               slMsgCopyT = 0;      // number of messages copied
   _U08              ubFifoStatusT;       // status of receive FIFO      
   _TsCpMsgQueue *   ptsQueueT;           // pointer to CAN message queue
   _TsCpCanMsg *     ptsCanMsgT;          // copy of pointer to CAN message
   
   
   //----------------------------------------------------------------
   // test the _TsCpPort pointer
   //
   if(ptsCanPortV == 0L)
   {
      return(-EINVAL);  
   }
   
   
   //----------------------------------------------------------------
   // test the logical interface number
   //
   slLogIfT = ptsCanPortV->slLogIf;
   if( (slLogIfT > CP_CHANNEL_MAX - 1) || (slLogIfT < 0) )
   {
      PK_DBG(2, "No CAN interface on port %d", slLogIfT);
      return(-ENODEV);
   }
   
   can_print_frame(ptsCanMsgV, szMsgDbgS);
   PK_DBG(eDBG_FUNC, "%s", szMsgDbgS);

   //----------------------------------------------------------------
   // run through all open message queues and push message into
   // the queue
   //
   for(slQueueCntT = 0; slQueueCntT < CP_QUEUE_MAX; slQueueCntT++)
   {
      //---------------------------------------------------
      // select the next possible open rcv queue
      //
      ptsQueueT = &(atsCpHalS[slLogIfT].atsMsgQueue[slQueueCntT]);
      if(ptsQueueT->tsCanPort.slQueue < 0) continue;

      //---------------------------------------------------
      // copy data into rcv queue 
      //
      ptsCanMsgT = ptsCanMsgV;
      for(slMsgCopyT = 0; slMsgCopyT < slMsgNumV; slMsgCopyT++)
      {
         ubFifoStatusT = CpFifoPush( &(ptsQueueT->tsRcvFifo), ptsCanMsgV);
         if(ubFifoStatusT == CpErr_FIFO_FULL) break;
         ptsCanMsgT++; // increase copy of CAN message ptr
      }
      
   }
   
   return(slMsgCopyT);  
}


//----------------------------------------------------------------------------//
// can_queue_trm_pop()                                                        //
// read CAN messages from Transmit queue ( all open queues)                   //
//----------------------------------------------------------------------------//
int can_queue_trm_pop(  _TsCpPort * ptsCanPortV, 
                        _TsCpCanMsg * ptsCanMsgV, int slMsgNumV)
{
   int               slLogIfT;            // number of logical CAN interface   
   int               slQueueCntT;         // counter for message queue
   int               slMsgCopyT = 0;      // number of messages copied
   _U08              ubFifoStatusT;       // status of receive FIFO      
   _TsCpMsgQueue *   ptsQueueT;           // pointer to CAN message queue
   
   
   //----------------------------------------------------------------
   // test the _TsCpPort pointer
   //
   if(ptsCanPortV == 0L)
   {
      return(-EINVAL);  
   }
   
   
   //----------------------------------------------------------------
   // test the logical interface number
   //
   slLogIfT = ptsCanPortV->slLogIf;
   if( (slLogIfT > CP_CHANNEL_MAX - 1) || (slLogIfT < 0) )
   {
      PK_DBG(2, "No CAN interface on port %d", slLogIfT);
      return(-ENODEV);
   }

   //----------------------------------------------------------------
   // run through all open message queues and push message into
   // the queue
   //
   slMsgCopyT = 0;
   for(slQueueCntT = 0; slQueueCntT < CP_QUEUE_MAX; slQueueCntT++)
   {
      //---------------------------------------------------
      // select the next possible open trm queue
      //
      ptsQueueT = &(atsCpHalS[slLogIfT].atsMsgQueue[slQueueCntT]);
      if(ptsQueueT->tsCanPort.slQueue < 0) continue;

      //---------------------------------------------------
      // copy data from trm queue 
      //
      ubFifoStatusT = CpFifoPop( &(ptsQueueT->tsTrmFifo), ptsCanMsgV);
      while(ubFifoStatusT != CpErr_FIFO_EMPTY)
      {
         ptsCanMsgV++;  // increase CAN message ptr
         slMsgCopyT++;  // increase message counter
         if(slMsgCopyT == slMsgNumV) return(slMsgCopyT);

         ubFifoStatusT = CpFifoPop( &(ptsQueueT->tsTrmFifo), ptsCanMsgV);
      };

   }
   
   return(slMsgCopyT);  
}


//----------------------------------------------------------------------------//
// can_queue_trm_push()                                                       //
// write CAN messages to specified CAN port (queue)                           //
//----------------------------------------------------------------------------//
int can_queue_trm_push( _TsCpPort * ptsCanPortV, 
                        _TsCpCanMsg * ptsCanMsgV, int slMsgNumV)
{
   int               slLogIfT;            // number of logical CAN interface   
   int               slQueueIdT = -1;     // selected queue ID
   int               slMsgCopyT;          // number of messages copied
   _U08              ubFifoStatusT;       // status of receive FIFO      
   _TsCpMsgQueue *   ptsQueueT;           // pointer to CAN message queue
   
   
   //----------------------------------------------------------------
   // test the _TsCpPort pointer
   //
   if(ptsCanPortV == 0L)
   {
      return(-EINVAL);  
   }
   
   
   //----------------------------------------------------------------
   // test the logical interface number
   //
   slLogIfT = ptsCanPortV->slLogIf;
   if( (slLogIfT > CP_CHANNEL_MAX - 1) || (slLogIfT < 0) )
   {
      PK_DBG(2, "No CAN interface on port %d", slLogIfT);
      return(-ENODEV);
   }

   //----------------------------------------------------------------
   // test the queue ID number
   //
   slQueueIdT = ptsCanPortV->slQueue;
   if( (slQueueIdT < 0) || (slQueueIdT > CP_QUEUE_MAX - 1) )
   {
      PK_DBG(2, "No queue ID %d on CAN interface %d", slQueueIdT, slLogIfT);
      return(-EINVAL);  
   }

      PK_DBG(eDBG_SOCK, "Message: ID=%04X DLC=%1d Data=%02X %02X", 
             ptsCanMsgV->tuMsgId.ulExt, ptsCanMsgV->ubMsgDLC, 
             ptsCanMsgV->tuMsgData.aubByte[0],
             ptsCanMsgV->tuMsgData.aubByte[1]);
   
   //----------------------------------------------------------------
   // select the message queue and write 'slMsgNumV' messages
   //
   ptsQueueT = &(atsCpHalS[slLogIfT].atsMsgQueue[slQueueIdT]);
   for(slMsgCopyT = 0; slMsgCopyT < slMsgNumV; slMsgCopyT++)
   {
      ubFifoStatusT = CpFifoPush(&(ptsQueueT->tsTrmFifo), ptsCanMsgV);
      if(ubFifoStatusT == CpErr_FIFO_FULL) break;
      ptsCanMsgV++;
   }
   
   return(slMsgCopyT);  
}



//============================================================================//

//----------------------------------------------------------------------------//
// can_clear_hal_struct()                                                     //
//                                                                            //
//----------------------------------------------------------------------------//
static void can_clear_hal_struct(int slLogIfV)
{
   int   slQueueCntT;
   
  
   //----------------------------------------------------------------
   // check if the logical interface number is in range
   //   
   if( (slLogIfV > CP_CHANNEL_MAX - 1) || (slLogIfV < 0) )
   {
      return;
   }
   
   //----------------------------------------------------------------
   // clear the complete HAL structure for that logic interface
   //
   memset(&atsCpHalS[slLogIfV], 0x00, sizeof(_TsCpHal));
   
   //----------------------------------------------------------------
   // interface not assigned yet
   //
   atsCpHalS[slLogIfV].tsCanPort.slLogIf = -1;
   atsCpHalS[slLogIfV].tsCanPort.slPhyIf = -1;
   
   atsCpHalS[slLogIfV].slStatus = CP_HAL_DRV_NONE;
   
   //----------------------------------------------------------------
   // mark all message queues as not assigned
   //
   for(slQueueCntT = 0; slQueueCntT < CP_QUEUE_MAX; slQueueCntT++)
   {
      atsCpHalS[slLogIfV].atsMsgQueue[slQueueCntT].tsCanPort.slQueue = -1;
   }
}


//----------------------------------------------------------------------------//
// can_print_frame()                                                          //
// print CAN frame to buffer                                                  //
//----------------------------------------------------------------------------//
#ifdef  CP_DEBUG
static void can_print_frame(_TsCpCanMsg * ptsCanMsgV, char * szBufferV)
{
   int   slPosT;
   _U08  ubDataCntT;
   _U08  ubDlcT;   
   
 
   //----------------------------------------------------------------
   // Format is:
   // ID(e): ----- DLC: - Data: -- -- .. -- (data frame,  Ext. ID)
   // ID(s):   --- DLC: - Data: -- -- .. -- (data frame,  Std. ID)
   // ID(s):   --- DLC: - RTR               (remote frame,  Std. ID)
   // DLC is set to x if value > 8
   //
   if(CpMsgIsExtended(ptsCanMsgV))
   {
      slPosT = sprintf( &szBufferV[0], "ID(e): 0x%08X ",
                        CpMsgGetExtId(ptsCanMsgV));
   }
   else
   {
      slPosT = sprintf( &szBufferV[0], "ID(s):      0x%03X ",
                        CpMsgGetStdId(ptsCanMsgV));
   }

   //----------------------------------------------------------------
   //  print DLC
   //  
   ubDlcT = CpMsgGetDlc(ptsCanMsgV);
   if(ubDlcT > 8)
   {
      slPosT += sprintf( &szBufferV[slPosT], "DLC: x ");
      ubDlcT = 8; // limit to highest possible value
   }
   else
   {
      slPosT += sprintf( &szBufferV[slPosT], "DLC: %d ", ubDlcT);
   }
   
   //----------------------------------------------------------------
   // check for remote frame
   //   
   if(CpMsgIsRemote(ptsCanMsgV))
   {
      sprintf( &szBufferV[slPosT], "RTR");
   }
   else
   {
      slPosT += sprintf( &szBufferV[slPosT], "Data: ");
      for(ubDataCntT = 0; ubDataCntT < ubDlcT; ubDataCntT++)
      {
         slPosT += sprintf( &szBufferV[slPosT], "%02X ", 
                           CpMsgGetData(ptsCanMsgV,ubDataCntT));
      }
   }
}
#endif


//----------------------------------------------------------------------------//
// can_register_driver()                                                      //
//                                                                            //
//----------------------------------------------------------------------------//
int can_register_driver(int slLogIfV, int slPhyIfV, _TsCpHal * ptsCpHalV)
{
   int         slCanIfIdxT;   // iterate over all CAN interface numbers
   int         slCanIfSelT;   // assigned logic CAN interface number
   int         slQueueCntT;   // counter for message queues
   _TsCpHal *  ptsHalT;       // pointer to HAL entry
   
      
   //----------------------------------------------------------------
   // test the logic interface number
   //
   if( slLogIfV >= CP_CHANNEL_MAX) 
   {
      PK_INF("Log. CAN-I/F number out of range (MAX=%d)", (CP_CHANNEL_MAX -1));
      return(-CpErr_CHANNEL);
   }

   //----------------------------------------------------------------
   // test the physical interface number
   //
   if( slPhyIfV >= CP_CHANNEL_MAX) 
   {
      PK_INF("Phy. CAN-I/F number out of range (MAX=%d)", (CP_CHANNEL_MAX -1));
      return(-CpErr_CHANNEL);
   }
   
   PK_DBG(eDBG_PORT, "register log. CAN-I/F %d", slLogIfV);

   //----------------------------------------------------------------
   // test the _TsCpHal structure
   //
   if( ptsCpHalV == 0L) return(-CpErr_GENERIC);

   
   //-----------------------------------------------------------------
   // does the driver request the next free port?
   //
   if(slLogIfV < 0)
   {
      ptsHalT = 0L;                 // no free slot yet
      slCanIfIdxT = CP_CHANNEL_1;   // start at first port
      
      //---------------------------------------------------------
      // search in all possible entries for a free slot
      //
      while(slCanIfIdxT < CP_CHANNEL_MAX)
      {
         //-------------------------------------------------
         // the port is not assigned if the structure
         // element 'slStatus' is CP_HAL_DRV_NONE
         //
         if(atsCpHalS[slCanIfIdxT].slStatus == CP_HAL_DRV_NONE)
         {
            //-----------------------------------------
            // we found a free slot here
            //
            ptsHalT = &(atsCpHalS[slCanIfIdxT]);
            break;
         }
         slCanIfIdxT++;
      }
   
      //---------------------------------------------------------
      // If 'ptsHalT' is still set to 0L, we could not find
      // a free slot for the driver
      //
      if(ptsHalT == 0L) 
      {
         PK_INF("No free log. CAN-I/F available");
         return(-CpErr_CHANNEL);
      }
      slCanIfSelT = slCanIfIdxT;
   }
   else
   {
      //----------------------------------------------------
      // At this position we try to assign the driver at a 
      // specific slot, given by the index 'slLogIfV'. We 
      // have to check if the requested port is free for 
      // allocation.
      //
      slCanIfSelT = slLogIfV;
      if(atsCpHalS[slCanIfSelT].slStatus != CP_HAL_DRV_NONE)
      {
         PK_INF("Log. CAN-/IF %d already assigned", slCanIfSelT);
         return(-CpErr_CHANNEL);
      }
      ptsHalT = &(atsCpHalS[slCanIfSelT]);
   }

   
   //----------------------------------------------------------------
   // copy the HAL structure from the driver and store
   // the assigned port number
   //
   memcpy(ptsHalT, ptsCpHalV, sizeof(_TsCpHal));
   ptsHalT->slStatus = CP_HAL_DRV_OK;
   ptsHalT->tsCanPort.slLogIf = slCanIfSelT;
   ptsHalT->tsCanPort.slPhyIf = slPhyIfV;
   
   
   //----------------------------------------------------------------
   // mark all message queues as not assigned
   //
   for(slQueueCntT = 0; slQueueCntT < CP_QUEUE_MAX; slQueueCntT++)
   {
      ptsHalT->atsMsgQueue[slQueueCntT].tsCanPort.slQueue = -1;
   }

   
   //----------------------------------------------------------------
   // add this device to the /proc interface
   //
   can_proc_dev_add(slCanIfSelT);
   
   PK_INF("Log. CAN-I/F %d has been assigned", slCanIfSelT);  
   return(slCanIfSelT);
}


//----------------------------------------------------------------------------//
// can_unregister_driver()                                                    //
//                                                                            //
//----------------------------------------------------------------------------//
void can_unregister_driver(int slLogIfV)
{
   //----------------------------------------------------------------
   // test the port number
   //
   if( slLogIfV >= CP_CHANNEL_MAX) return;
   if( slLogIfV <  0) return;

   PK_DBG(eDBG_PORT, "unregister log. CAN-I/F %d", slLogIfV);
   can_proc_dev_remove(slLogIfV);
   can_clear_hal_struct(slLogIfV);
}


//----------------------------------------------------------------------------//
// can_get_hal()                                                              //
//                                                                            //
//----------------------------------------------------------------------------//
_TsCpHal * can_get_hal(int slLogIfV)
{
   //----------------------------------------------------------------
   // test the port number
   //
   if( slLogIfV >= CP_CHANNEL_MAX) 
   {
      return(0L);
   }

   return( &(atsCpHalS[slLogIfV]) );
}

//----------------------------------------------------------------------------//
// can_get_port()                                                             //
//                                                                            //
//----------------------------------------------------------------------------//
_TsCpPort * can_get_port(int slLogIfV)
{
   //----------------------------------------------------------------
   // test the port number
   //
   if( slLogIfV >= CP_CHANNEL_MAX) 
   {
      return(0L);
   }

   return( &(atsCpHalS[slLogIfV].tsCanPort) );
}




//----------------------------------------------------------------------------//
// init_module()                                                              //
//
//----------------------------------------------------------------------------//
static int __init ModInit(void)
{
   _U08  ubCanIfNumT;

   //----------------------------------------------------------------
   // print some info who we are
   //       
   PK_INF("CANpie version %s", CP_MODULE_VERSION);
   PK_INF("Starting kernel driver");
   #ifdef  CP_DEBUG
   PK_INF("Debug support enabled, debug level=%d", uwCpDebugG);
   #endif
   
   //----------------------------------------------------------------
   // clear the structure for the CAN Hardware Abstraction layer
   //
   for(ubCanIfNumT = 0; ubCanIfNumT < CP_CHANNEL_MAX; ubCanIfNumT++)
   {
      can_clear_hal_struct(ubCanIfNumT);
   }

   //-----------------------------------------------------------------
   // socket support (AF_CAN)
   //
   #if CP_CONFIG_SOCKET == 1
   can_socket_init();
   #endif
   
   
   //-----------------------------------------------------------------
   // support for /dev/can character device driver
   //
   #if CP_CONFIG_DEV == 1
   cp_dev_init();
   #endif
   
   //-----------------------------------------------------------------
   // support for /proc filesystem
   //
   can_proc_init();
   
   return(0);
}


//----------------------------------------------------------------------------//
// cleanup_module()                                                           //
//
//----------------------------------------------------------------------------//
static void __exit ModExit(void)
{

   //-----------------------------------------------------------------
   // socket support (AF_CAN)
   //
   #if CP_CONFIG_SOCKET == 1
   can_socket_cleanup();
   #endif
   
   
   //-----------------------------------------------------------------
   // support for /dev/can character device driver
   //
   #if CP_CONFIG_DEV == 1
   cp_dev_cleanup();
   #endif
   

   //-----------------------------------------------------------------
   // remove /proc/net/can filesystem
   //
   can_proc_cleanup();
         
   
   PK_INF("CANpie Kernel driver removed");
}


module_init( ModInit );
module_exit( ModExit );
