//****************************************************************************//
// File:          canpie_proc.c                                               //
// Description:   CANpie /proc support for Linux                              //
// Author:        Uwe Koppe                                                   //
// e-mail:        koppe@microcontrol.net                                      //
//                                                                            //
//============================================================================//
// This program is free software; you can redistribute it and/or modify       //
// it under the terms of the GNU Lesser General Public License as published   //
// by the Free Software Foundation; either version 2.1 of the License, or     //
// (at your option) any later version.                                        //
//============================================================================//
//                                                                            //
// Date        History                                                        //
// ----------  -------------------------------------------------------------- //
// 29.07.2003  Initial version                                                //
//                                                                            //
//****************************************************************************//


//------------------------------------------------------------------------------
// CVS version information:
// $Id: canpie_proc.c,v 1.2 2005/06/17 13:48:37 microcontrol Exp $
//------------------------------------------------------------------------------


/*----------------------------------------------------------------------------*\
** Include files                                                              **
**                                                                            **
\*----------------------------------------------------------------------------*/

#include <linux/proc_fs.h>
#include <linux/kernel.h>
#include <linux/module.h>

#include <net/can/canpie_dbg.h>
#include <net/can/canpie_hal.h>
#include <net/can/cpcore.h>


/*----------------------------------------------------------------------------*\
** Definitions                                                                **
**                                                                            **
\*----------------------------------------------------------------------------*/







/*----------------------------------------------------------------------------*\
** /proc filesystem support                                                   **
** Static variables and function prototypes                                   **
\*----------------------------------------------------------------------------*/



#ifdef CONFIG_PROC_FS



static int slProcCan_InitializedS = 0;


int         can_proc_init(void);
int         can_proc_cleanup(void);

int         can_proc_dev_add(int slCanIfV);
int         can_proc_dev_remove(int slCanIfV);


//----------------------------------------------------------------------------//
// Internal (static) functions for access to the files in the                 //
// directory '/proc/net/can'.                                                 //
// Functions for read access have the suffix '_rd'                            //
// Functions for write access have the suffix '_wr'                           //
//                                                                            //
//----------------------------------------------------------------------------//

//-------------------------------------------------------------------
// file  : /proc/net/can/info
// access: read-only
// 
static int  proc_can_info_rd     (  char *page, char **start, off_t off,
                                    int count, int *eof , void *data       );

//-------------------------------------------------------------------
// file  : /proc/net/can/debug
// access: read/write
// info  : modify the debug level of canpie module
//
static int  proc_can_debug_rd    (  char *page, char **start, off_t off,
                                    int count, int *eof , void *data       );

static int  proc_can_debug_wr    (  struct file *file, const char *buffer,
                                    unsigned long count, void *data        );
                               

//-------------------------------------------------------------------
// file  : /proc/net/can/canX/baud
// access: read/write
// info  : modify the baudrate of the CAN interface
//
static int  proc_dev_baud_rd     (  char *page, char **start, off_t off,
                                    int count, int *eof , void *data       );

static int  proc_dev_baud_wr     (  struct file *file, const char *buffer,
                                    unsigned long count, void *data        );

                                                                        
//-------------------------------------------------------------------
// file  : /proc/net/can/canX/mode
// access: read/write
// info  : modify the mode of the CAN interface (start/stop)
//
static int  proc_dev_mode_rd     (  char *page, char **start, off_t off,
                                    int count, int *eof , void *data       );

static int  proc_dev_mode_wr     (  struct file *file, const char *buffer,
                                    unsigned long count, void *data        );
                                                                     
static struct proc_dir_entry *   ptsProcCanRootS;

static struct proc_dir_entry *   atsProcCanDevS[CP_CHANNEL_MAX];

#else /* undef CONFIG_PROC_FS */


#define  can_proc_init()         0
#define  can_proc_cleanup()      0

#define  can_proc_dev_add()      0
#define  can_proc_dev_remove()   0


#endif   // CONFIG_PROC_FS



/*----------------------------------------------------------------------------*\
** Module variables                                                           **
**                                                                            **
\*----------------------------------------------------------------------------*/



/*----------------------------------------------------------------------------*\
** Functions                                                                  **
**                                                                            **
\*----------------------------------------------------------------------------*/




/*----------------------------------------------------------------------------*\
** /proc filesystem support                                                   **
** function implementation                                                    **
\*----------------------------------------------------------------------------*/



#ifdef CONFIG_PROC_FS

//----------------------------------------------------------------------------//
// can_proc_dev_add()                                                         //
//                                                                            //
//----------------------------------------------------------------------------//
int can_proc_dev_add(int slCanIfV)
{
   struct proc_dir_entry *  ptsProcCanEntryT;
   char                     szPortNameT[16];


   //-----------------------------------------------------------------
   // check if the supplied port number is in range and not
   // occupied yet
   //
   if( (slCanIfV < 0) || (slCanIfV >= CP_CHANNEL_MAX) )
   {
      PK_ERR("Can't create /proc/net/can/ entry for port %d", slCanIfV);
      return(-EPERM);
   }

   if(atsProcCanDevS[slCanIfV] != 0L)
   {
      PK_ERR("Directory /proc/net/can/can%d already assigned", slCanIfV);
      return(-EEXIST);
   }

   if(ptsProcCanRootS == 0L)
   {
      PK_ERR("/proc/net/can/ does not exist");
      return(-ENOENT);
   }

   //-----------------------------------------------------------------
   // create the directory for canN
   //
   sprintf(szPortNameT, "can%d", slCanIfV); 
   atsProcCanDevS[slCanIfV] = proc_mkdir(szPortNameT, ptsProcCanRootS);
   if(atsProcCanDevS[slCanIfV] == 0L)
   {
      PK_ERR("Could not create /proc/net/can/%s", szPortNameT);
      return(-ENOENT);
   }


   //-----------------------------------------------------------------
   // the 'baud' file holds the current buadrate
   //
   ptsProcCanEntryT = create_proc_entry("baud", 
                                        S_IFREG | S_IRUGO | S_IWUSR, 
                                        atsProcCanDevS[slCanIfV]);
   if(ptsProcCanEntryT)
   {
      ptsProcCanEntryT->read_proc   = &proc_dev_baud_rd;
      ptsProcCanEntryT->write_proc  = &proc_dev_baud_wr;
      ptsProcCanEntryT->owner       = THIS_MODULE;
      ptsProcCanEntryT->data        = can_get_port(slCanIfV);
   }

   //-----------------------------------------------------------------
   // the 'mode' file holds the CAN mode (stop/start)
   //
   ptsProcCanEntryT = create_proc_entry("mode", 
                                        S_IFREG | S_IRUGO | S_IWUSR, 
                                        atsProcCanDevS[slCanIfV]);
   if(ptsProcCanEntryT)
   {
      ptsProcCanEntryT->read_proc   = &proc_dev_mode_rd;
      ptsProcCanEntryT->write_proc  = &proc_dev_mode_wr;
      ptsProcCanEntryT->owner       = THIS_MODULE;
      ptsProcCanEntryT->data        = can_get_port(slCanIfV);
   }
      
   return(0);
}


//----------------------------------------------------------------------------//
// cp_proc_candev_del()                                                       //
//                                                                            //
//----------------------------------------------------------------------------//
int can_proc_dev_remove(int slCanIfV)
{
   char   szPortNameT[16];

   //----------------------------------------------------------------
   // create the name for the sub-directory of /proc/net/can
   //
   sprintf(szPortNameT, "can%d", slCanIfV); 

   //----------------------------------------------------------------
   // test if there is a proc-entry available
   //
   if(ptsProcCanRootS == 0L)
   {
      PK_ERR("/proc/net/can/ does not exist");
      return(-ENOENT);
   }
   
   if(atsProcCanDevS[slCanIfV] == 0L)
   {
      PK_ERR("/proc/net/can/%s does not exist", szPortNameT);
      return(-ENOENT);
   }


   //-----------------------------------------------------------------
   // remove the files in the directory /proc/net/can/canX
   //
   remove_proc_entry("baud", atsProcCanDevS[slCanIfV]);
   remove_proc_entry("mode", atsProcCanDevS[slCanIfV]);

      
   //-----------------------------------------------------------------
   // remove the sub-directory
   //
   remove_proc_entry(szPortNameT, ptsProcCanRootS);

   return(0);
}


//----------------------------------------------------------------------------//
// can_proc_init()                                                            //
//                                                                            //
//----------------------------------------------------------------------------//
int can_proc_init(void)
{
   struct proc_dir_entry *    ptsProcCanEntryT;
   int                        slDevIndexT;

   //----------------------------------------------------------------
   // create the proc filesystem for CAN
   // it is located in "/proc/net/"
   //
   slProcCan_InitializedS = 0;


   if(! proc_net)
   {
      PK_ERR("/proc/net/ does not exist");
      can_proc_cleanup();
      return -ENOENT;
   }


   ptsProcCanRootS = proc_mkdir("can", proc_net);
   if(!ptsProcCanRootS)
   {
      PK_ERR("Could not create /proc/net/can/");
      return -ENOENT;
   }


   //-----------------------------------------------------------------
   // the 'info' file holds the CANpie version and information
   // about the core features (e.g. number of ports, etc.)
   //
   ptsProcCanEntryT = create_proc_entry("info", 
                                        S_IFREG | S_IRUGO, 
                                        ptsProcCanRootS);
   if(ptsProcCanEntryT)
   {
      ptsProcCanEntryT->read_proc   = &proc_can_info_rd;
      ptsProcCanEntryT->write_proc  = NULL;
      ptsProcCanEntryT->owner       = THIS_MODULE;
   }

   //-----------------------------------------------------------------
   // the 'debug' file allows modification of the debug level
   //
   ptsProcCanEntryT = create_proc_entry("debug", 
                                        S_IFREG | S_IRUGO | S_IWUSR, 
                                        ptsProcCanRootS);
   if(ptsProcCanEntryT)
   {
      ptsProcCanEntryT->read_proc   = &proc_can_debug_rd;
      ptsProcCanEntryT->write_proc  = &proc_can_debug_wr;
      ptsProcCanEntryT->owner       = THIS_MODULE;
   }
   


   //-----------------------------------------------------------------
   // clear the array 'atsProcCanDevS', since these entries
   // are not created yet
   //
   for(slDevIndexT = 0; slDevIndexT < CP_CHANNEL_MAX; slDevIndexT++)
   {
      atsProcCanDevS[slDevIndexT] = NULL;
   }

   slProcCan_InitializedS = 1;
   PK_INF("/proc/net/can initialized");
   return 0;
}


//----------------------------------------------------------------------------//
// can_proc_cleanup()                                                         //
//                                                                            //
//----------------------------------------------------------------------------//
int can_proc_cleanup(void)
{
   if(slProcCan_InitializedS > 0)
   {
      remove_proc_entry("debug", ptsProcCanRootS);
      remove_proc_entry("info" , ptsProcCanRootS);
      
      remove_proc_entry("can"  , proc_net);
      PK_INF("/proc/net/can removed");
      slProcCan_InitializedS = 0;
   }
   return 0;
}





//----------------------------------------------------------------------------//
// proc_can_debug_rd()                                                        //
//                                                                            //
//----------------------------------------------------------------------------//
int proc_can_debug_rd(  char *page, char **start, off_t off,
                        int count, int *eof , void *data )
{
   int   size = 0;
   int   slDebugLevel = 0;   
   
   #ifdef  CP_DEBUG
   slDebugLevel = uwCpDebugG;
   #endif
   size += sprintf(page + size, "%d\n", slDebugLevel);
   
   return(size);
}


//----------------------------------------------------------------------------//
// proc_can_debug_wr()                                                        //
//                                                                            //
//----------------------------------------------------------------------------//
int proc_can_debug_wr(  struct file *file, const char *buffer,
                        unsigned long count, void *data     )
{
   int   slNewDebugLevelT = 0;
   
   //----------------------------------------------------------------
   // brute force atoi operation
   //
   slNewDebugLevelT = (int)(*buffer - '0');
   if(slNewDebugLevelT < 0)
   {
      slNewDebugLevelT = 0;   
   }
   
   if(slNewDebugLevelT > eDBG_MAX)
   {
      slNewDebugLevelT = eDBG_MAX;
   }
   
   #ifdef  CP_DEBUG
   uwCpDebugG = slNewDebugLevelT;
   #endif
   
   return(count);
}

//----------------------------------------------------------------------------//
// proc_can_info_rd()                                                         //
//                                                                            //
//----------------------------------------------------------------------------//
int proc_can_info_rd(char *page, char **start, off_t off,
                     int count, int *eof , void *data )
{
   int    size = 0;
   char   szMode[8];
   
   size += sprintf(page + size, 
                   "CANpie version          : %s \n"   , CP_MODULE_VERSION);
                   
   size += sprintf(page + size, 
                   "Logic port number range : 0..%d \n", CP_CHANNEL_MAX - 1);
   
   #if CP_CONFIG_DEV==1
   sprintf(szMode, "Yes");
   #else
   sprintf(szMode, "No");
   #endif
   size += sprintf(page + size, 
                   "Support /dev/can filesystem: %s \n", szMode);
   return(size);
}


//----------------------------------------------------------------------------//
// proc_dev_baud_rd()                                                         //
//                                                                            //
//----------------------------------------------------------------------------//
int proc_dev_baud_rd(char *page, char **start, off_t off,
                     int count, int *eof , void *data )
{
   int   size = 0;
   int   slBaudSelT = 0;   
   size += sprintf(page + size, "%d\n", slBaudSelT);
   
   return(size);
}


//----------------------------------------------------------------------------//
// proc_dev_baud_wr()                                                         //
//                                                                            //
//----------------------------------------------------------------------------//
int proc_dev_baud_wr(struct file *file, const char *buffer,
                     unsigned long count, void *data     )
{
   _TsCpBitrate   tsBitrateT;
   
   //----------------------------------------------------------------
   // brute force atoi operation
   //
   tsBitrateT.ubBaudSel = (unsigned char)(*buffer - '0');
   
   if(tsBitrateT.ubBaudSel >= CP_BAUD_BTR)
   {
      tsBitrateT.ubBaudSel = CP_BAUD_BTR - 1;
   }
   
   CpCoreBitrate(data,&tsBitrateT);
   
   return(count);
}


//----------------------------------------------------------------------------//
// proc_dev_mode_rd()                                                         //
//                                                                            //
//----------------------------------------------------------------------------//
int proc_dev_mode_rd(char *page, char **start, off_t off,
                     int count, int *eof , void *data )
{
   int   size = 0;
   _U08  ubStatusT;
   
   CpCoreCanStatus(data, &ubStatusT);
   
   size += sprintf(page + size, "%d\n", ubStatusT);
   
   return(size);
}


//----------------------------------------------------------------------------//
// proc_dev_baud_wr()                                                         //
//                                                                            //
//----------------------------------------------------------------------------//
int proc_dev_mode_wr(struct file *file, const char *buffer,
                     unsigned long count, void *data     )
{
   _U08  ubModeSelT = 0;
      
   //----------------------------------------------------------------
   // brute force atoi operation
   //
   ubModeSelT = (unsigned char)(*buffer - '0');
   
   if(ubModeSelT > CP_MODE_START)
   {
      ubModeSelT = CP_MODE_START;
   }
   
   CpCoreCanMode(data, ubModeSelT);
   
   return(count);
}

#endif   // CONFIG_PROC_FS
