/*****************************************************************************\
*  CANpie                                                                     *
*                                                                             *
*  File        : mc_pcmcia.c                                                  *
*  Description : Kernel driver for the MicroControl mCAN.pc.PCMCIA            *
*  Author      : Uwe Koppe                                                    *
*  e-mail      : koppe@microcontrol.net                                       *
*                                                                             *
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - *
*                                                                             *
*   This program is free software; you can redistribute it and/or modify      *
*   it under the terms of the GNU General Public License as published by      *
*   the Free Software Foundation; either version 2 of the License, or         *
*   (at your option) any later version.                                       *
*                                                                             *
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - *
*                                                                             *
*                                                                             *
\*****************************************************************************/

#include <asm/atomic.h>          // atomic data types and operations
#include <asm/delay.h>           // delay()
#include <asm/io.h>              // inb() and outb()
#include <linux/timer.h>         // add_timer()
#include <linux/sched.h>         // schedule()

#include <net/can/canpie.h>
#include <net/can/canpie_hal.h>
#include <net/can/cpcanmsg.h>

#include "mc_pcmcia_cs.h"



/*------------------------------------------------------------------------
** Global variables used for the core functions
**
*/


/*------------------------------------------------------------------------
** Global variables used for access to the CAN card
**
*/
#define  PCMCIA_BUFFER_SIZE      16
#define  PCMCIA_CHANNEL_MAX       2
#define  PCMCIA_COM_SIZE         16          // size of communication buffer
      

static _U08 aubDataReadBufG[PCMCIA_COM_SIZE];      // buffer for read operation
static _U08 aubDataWriteBufG[PCMCIA_COM_SIZE];     // buffer for write operation

static _TsMcPcmciaInfo *   ptsInfoS = NULL;
static struct timer_list   tsTimerS;
static int  aslCanLogIf[PCMCIA_CHANNEL_MAX];       // assignment of logical CAN-I/F

static atomic_t   ulTimerStopS;

//extern _U16 uwCpPcmciaInterruptG;

extern _TsCpCanMsg   tsCanRcvMsgG;
_TsCpCanMsg   tsCanTrmMsgG;


//-----------------------------------------------------------------------------
// the CpVar_InitFlag is used to keep track if a channel
// is already initialized,
// Bit 0 = channel 0
// Bit 1 = channel 1
// ....
// Bit 15 = channel 15
// a bit value of 0 means not initialized
// a bit value of 1 means initialized
_U16 CpVar_InitFlag = 0;


/*------------------------------------------------------------------------
** CpStruct_BitTimingTable
** The bit timing structure has to be fitted to the used CAN controller.
** If more than one CAN channel is used the structure has to be extended.
*/
const _TsCpBitrate atsBitTimingTable[] = {
   { 0x00, 0x31, 0x1C, 0x00 },   /*  10 KBit/sec   */
   { 0x01, 0x18, 0x1C, 0x01 },   /*  20 KBit/sec   */
   { 0x02, 0x09, 0x1C, 0x02 },   /*  50 KBit/sec   */
   { 0x03, 0x04, 0x1C, 0x03 },   /* 100 KBit/sec   */
   { 0x04, 0x03, 0x1C, 0x04 },   /* 125 KBit/sec   */
   { 0x05, 0x01, 0x1C, 0x05 },   /* 250 KBit/sec   */
   { 0x06, 0x00, 0x1C, 0x06 },   /* 500 KBit/sec   */
   { 0x07, 0x00, 0x16, 0x07 },   /* 800 KBit/sec   */
   { 0x08, 0x00, 0x14, 0x08 },   /*   1 MBit/sec   */

};



/*------------------------------------------------------------------------
** constants when reading the status of the CAN interface
**
*/
const _U08 FIFO_MASK_RCV_EMPTY = 0x01;
const _U08 FIFO_MASK_TRM_FULL  = 0x04;

enum  FifoStatus_e {
   eFIFO_STATUS_EMPTY = 0,    // no data in FIFO
   eFIFO_STATUS_DATA,         // data in FIFO available
   eFIFO_STATUS_FULL
};


static _U08 mc_pcmcia_data_read(void);
static _U08 mc_pcmcia_data_write(void);
static _U08 mc_pcmcia_status(_U08 ubFifoTypeV);
static void mc_pcmcia_task(unsigned long ulParamV);


/*------------------------------------------------------------------------
**	CpCoreAllocBuffer()
**
*/
#if   CP_FULL_CAN == 1
_U08 Cp_PREFIX CpCoreAllocBuffer (_U08 channel, CpStruct_CAN * msg,
                                  _U08 direction)
{
#if   CP_SMALL_CODE == 0
   //---	test the channel number ------------------------------------
	if( (channel + 1) > CP_CHANNEL_MAX) return (CpErr_CHANNEL);
#endif

   return (CpErr_OK);
}
#endif


//----------------------------------------------------------------------------//
// CpPcmciaBitrate()                                                          //
// setup baudrate                                                             //
//----------------------------------------------------------------------------//
_TvCpStatus CpPcmciaBitrate(_TsCpPort * ptsPortV, _TsCpBitrate * ptsBitrateV)
{
   _U16  uwCountT;
   _U08  ubCanBtr0T;       // value written in btr0 register
   _U08  ubCanBtr1T;       // value written in btr1 register
   _U08  ubCanNumT;        // CAN number for PCMCIA interface
   
   
   //----------------------------------------------------------------
   // test the physical channel number
   //
   if( ptsPortV->slPhyIf >= PCMCIA_CHANNEL_MAX) return (CpErr_CHANNEL);

   //----------------------------------------------------------------
   // setup the values for btr0 and btr1
   //
   if(ptsBitrateV->ubBaudSel == CP_BAUD_BTR)
   {
      ubCanBtr0T = ptsBitrateV->ubBtr0;
      ubCanBtr1T = ptsBitrateV->ubBtr1;
   }
   else
   {
      //--- take values from table -----------------------------
      ubCanBtr0T = atsBitTimingTable[ptsBitrateV->ubBaudSel].ubBtr0;
      ubCanBtr1T = atsBitTimingTable[ptsBitrateV->ubBaudSel].ubBtr1;
   }


   PK_DBG("btr0=%X btr1=%X\n", ubCanBtr0T, ubCanBtr1T);

   //----------------------------------------------------------------
   // init the CAN port and setup baudrate
   //

   ubCanNumT = (_U08) (ptsPortV->slPhyIf) + 1;
   
   //--- setup the data in the write buffer ---------------
   aubDataWriteBufG[0] = 0x30;
   aubDataWriteBufG[1] = ubCanNumT;
   aubDataWriteBufG[2] = 0x00;
   aubDataWriteBufG[3] = 0x00;
   aubDataWriteBufG[4] = 0x00;
   aubDataWriteBufG[5] = 0x00;
   aubDataWriteBufG[6] = ubCanBtr0T;
   aubDataWriteBufG[7] = ubCanBtr1T;
   aubDataWriteBufG[8] = 0x00;
   aubDataWriteBufG[9] = 0x00;

   aubDataWriteBufG[10] = 0x00;
   aubDataWriteBufG[11] = 0x00;
   aubDataWriteBufG[12] = 0xFF;
   aubDataWriteBufG[13] = 0xE0;
   aubDataWriteBufG[14] = 0x00;
   aubDataWriteBufG[15] = 0x00;

   //--- send the data to the CAN card ------------------------------
   mc_pcmcia_data_write();

   //--- wait for 500 ms --------------------------------------------
   for(uwCountT = 0; uwCountT < 500; uwCountT++)
   {
      udelay(1000);
   }

   //--- read data from card ----------------------------------------
   mc_pcmcia_data_read();
   if( (aubDataReadBufG[0] == 0x00) && (aubDataReadBufG[1] == 0x01) )
   {
      PK_DBG("Baudrate setting OK\n");
      return (CpErr_OK);
   }

   PK_DBG("Baudrate setting failed\n");
   return (CpErr_BAUDRATE);
}


/*------------------------------------------------------------------------
** CpCoreBufferTransmit()
**
*/
#if   CP_FULL_CAN == 1
_U08 Cp_PREFIX CpCoreBufferTransmit(_U08 channel, CpStruct_CAN * msg)
{
#if   CP_SMALL_CODE == 0
   //--- test the channel number ------------------------------------
   if( (channel + 1) > CP_CHANNEL_MAX) return (CpErr_CHANNEL);
#endif

   return (CpErr_OK);
}
#endif


//----------------------------------------------------------------------------//
// CpPcmciaCanMode()                                                          //
// Set mode of CAN controller                                                 //
//----------------------------------------------------------------------------//
_TvCpStatus CpPcmciaCanMode(_TsCpPort * ptsPortV, _U08 ubModeV)
{
   _U08  ubReturnT = CpErr_OK;

   //----------------------------------------------------------------
   // test the channel number
   //

   return(ubReturnT);
}


//----------------------------------------------------------------------------//
// CpPcmciaCanStatus()                                                        //
// Retrive status of CAN controller                                           //
//----------------------------------------------------------------------------//
_TvCpStatus CpPcmciaCanStatus(_TsCpPort * ptsPortV, _U08 * pubStatusV)
{
   _U08  ubStatusT = 0;
   _U08  ubReturnT = CpErr_OK;

   //----------------------------------------------------------------
   // test the channel number
   //



   *pubStatusV = ubStatusT;
           
   return(ubReturnT);
}



//----------------------------------------------------------------------------//
// CpPcmciaDriverInit()                                                       //
// initialize the CAN card                                                    //
//----------------------------------------------------------------------------//
_TvCpStatus CpPcmciaDriverInit(_U08 ubPhysPortV, _TsCpPort * ptvPortV)
{

   //----------------------------------------------------------------


   return (CpErr_OK);
}


//----------------------------------------------------------------------------//
// CpPcmciaDriverRelease()                                                    //
// release the CAN driver                                                     //
//----------------------------------------------------------------------------//
_TvCpStatus CpPcmciaDriverRelease(_TsCpPort * ptvPortV)
{
   _U08  ubReturnT = CpErr_OK;




   return(ubReturnT);
}

/*------------------------------------------------------------------------
** CpCoreIntHandler()
**
*/
void CpCoreIntHandler(void)
{

}


//----------------------------------------------------------------------------//
// CpCoreHDI()                                                                //
//                                                                            //
//----------------------------------------------------------------------------//
_U08 CpCoreHDI(_U08 ubChannelV, CpStruct_HDI * hdi)
{
   //----------------------------------------------------------------
   // test the channel number
   //
   #if   CP_SMALL_CODE == 0
   if( (ubChannelV + 1) > CP_CHANNEL_MAX) return (CpErr_CHANNEL);
   #endif


   //----------------------------------------------------------------
   // set the support flags according to the configuration
   // options
   //

   return (CpErr_OK);
}


/*------------------------------------------------------------------------
** CpCoreFilterAll()
**
*/
_U08 Cp_PREFIX CpCoreFilterAll(_U08 channel, _BIT enable)
{
#if   CP_SMALL_CODE == 0
   //--- test the channel number ------------------------------------
   if( (channel + 1) > CP_CHANNEL_MAX) return (CpErr_CHANNEL);
#endif


   return (CpErr_OK);
}


/*------------------------------------------------------------------------
** CpCoreFilterMsg()
**
*/
_U08 Cp_PREFIX CpCoreFilterMsg(_U08 channel, _U16 id, _BIT enable)
{
#if   CP_SMALL_CODE == 0
   //--- test the channel number ------------------------------------
   if( (channel + 1) > CP_CHANNEL_MAX) return (CpErr_CHANNEL);
#endif


   return (CpErr_OK);
}


//----------------------------------------------------------------------------//
// CpPcmciaMsgRead()                                                          //
// Read CAN message from controller                                           //
//----------------------------------------------------------------------------//
_TvCpStatus CpPcmciaMsgRead(  _TsCpPort * ptsPortV, _TsCpCanMsg * ptsCanMsgV,
                              _U16 * puwMsgCntV)
{
   _U08  ubReturnT = CpErr_OK;
   _U16  uwMsgCntT = 0;
   

   //----------------------------------------------------------------
   // read data from the receive queue
   //
   uwMsgCntT = can_queue_rcv_pop(ptsPortV, ptsCanMsgV, *puwMsgCntV);
   *puwMsgCntV = uwMsgCntT;

   return(ubReturnT);   
}


//----------------------------------------------------------------------------//
// CpPcmciaMsgWrite()                                                         //
// send CAN message                                                           //
//----------------------------------------------------------------------------//
_TvCpStatus CpPcmciaMsgWrite( _TsCpPort * ptsPortV, _TsCpCanMsg * ptsCanMsgV,
                              _U16 * puwMsgCntV)
{
   _U08           ubTrmFifoStatusT;
   _U08           ubDataCntT;
   _U32           ulIdentifierT;
   _TsCpCanMsg    tsCanMsgT;
   
   
   //----------------------------------------------------------------
   // get status from PCMCIA transmit FIFO
   //
   ubTrmFifoStatusT = mc_pcmcia_status(CP_FIFO_TRM);
   
   //----------------------------------------------------------------
   // read as long from the transmit FIFO (memory) until the FIFO
   // on the PCMCIA card is full
   //
   while(ubTrmFifoStatusT != eFIFO_STATUS_FULL)
   {
      //---------------------------------------------------
      // copy data from the transmit queue
      //
      if( can_queue_trm_pop(ptsPortV, &tsCanMsgT, 1) != 1)
      {
         //--- no more messages to read --------------
         break;
      }
   
      //---------------------------------------------------
      // setup the data in the write buffer
      //
      if( CpMsgIsExtended(&tsCanMsgT) )
      {
         aubDataWriteBufG[0x00] = 0x45;            // command: extended message
         ulIdentifierT = CpMsgGetExtId(&tsCanMsgT);
      }
      else
      {
         aubDataWriteBufG[0x00] = 0x44;            // command: standard message
         ulIdentifierT = CpMsgGetStdId(&tsCanMsgT);
      }

      aubDataWriteBufG[0x01] = ptsPortV->slPhyIf + 1;     // channel

      aubDataWriteBufG[0x02] = CpMsgGetDlc(&tsCanMsgT);
      if(CpMsgIsRemote(&tsCanMsgT))
      {
         aubDataWriteBufG[0x02] |= 0x10;
      }
      
      //--- identifier --------------------------
      aubDataWriteBufG[0x03] = (_U08) ulIdentifierT;
      ulIdentifierT = ulIdentifierT >> 8;
      aubDataWriteBufG[0x04] = (_U08) ulIdentifierT;
      ulIdentifierT = ulIdentifierT >> 8;
      aubDataWriteBufG[0x05] = (_U08) ulIdentifierT;
      ulIdentifierT = ulIdentifierT >> 8;
      aubDataWriteBufG[0x07] = (_U08) ulIdentifierT;
      
      //--- data --------------------------------
      for(ubDataCntT = 7; ubDataCntT < 15; ubDataCntT++)
      {
         aubDataWriteBufG[ubDataCntT] = CpMsgGetData(&tsCanMsgT, ubDataCntT - 7);
      }

      //--- send the data to the CAN card ------------------------------
      mc_pcmcia_data_write();
      
   }

   
   
   return (CpErr_OK);
}



/*----------------------------------------------------------------------------*\
** module init and exit code                                                  **
** function implementation                                                    **
\*----------------------------------------------------------------------------*/


//----------------------------------------------------------------
// Setup the CpHal_s structure
//
static _TsCpHal tsMcPcmciaHalS = {
   pfnBitrate        : CpPcmciaBitrate,
   pfnBufferGetData  : 0L,
   pfnBufferGetDlc   : 0L,
   pfnBufferInit     : 0L,
   pfnBufferSetData  : 0L,
   pfnBufferSetDlc   : 0L,
   pfnBufferSend     : 0L,
   pfnBufferRelease  : 0L,
   pfnCanMode        : CpPcmciaCanMode,
   pfnCanStatus      : CpPcmciaCanStatus,
   pfnDriverInit     : CpPcmciaDriverInit,
   pfnDriverRelease  : CpPcmciaDriverRelease,
   pfnFilterAll      : 0L,
   pfnFilterMsg      : 0L,
   pfnHwDescription  : 0L,

   pfnMsgRead        : CpPcmciaMsgRead,
   pfnMsgWrite       : CpPcmciaMsgWrite,
   pfnRegRead        : 0L,
   pfnRegWrite       : 0L

};


//----------------------------------------------------------------------------//
// mc_pcmcia_can_register()                                                   //
// register the CAN driver to CANpie                                          //
//----------------------------------------------------------------------------//
int  mc_pcmcia_can_register(unsigned short * ptsPortArrayV)
{
   int   slRegisterT;
   int   slChannelCntT;
   
   //----------------------------------------------------------------
   // test the array for the port number assignment
   // it is initialized via "insmod <module> port=.."
   //
   if(ptsPortArrayV == NULL) 
   {
      return(-ENODEV);
   }

   //----------------------------------------------------------------
   // clear the logical CAN interface assignment
   //
   aslCanLogIf[1] = -1;
   
   
   //----------------------------------------------------------------
   // register the physical CAN interface on the requested
   // logical CAN interface number
   //
   for(slChannelCntT = 0; slChannelCntT < PCMCIA_CHANNEL_MAX; slChannelCntT++)
   {
      aslCanLogIf[slChannelCntT] = -1;
      slRegisterT = can_register_driver(  *ptsPortArrayV, 
                                          slChannelCntT, 
                                          &tsMcPcmciaHalS);
   
      if(slRegisterT < 0)
      {
         PK_ERR("registration of physical CAN-I/F %d failed", slChannelCntT);
      }
      else
      {
         aslCanLogIf[slChannelCntT] = slRegisterT;
      }
   
      ptsPortArrayV++;
   
   }
      
   return(0);
}


//----------------------------------------------------------------------------//
// mc_pcmcia_can_unregister()                                                 //
// unregister the CAN driver from CANpie                                      //
//----------------------------------------------------------------------------//
int  mc_pcmcia_can_unregister(void)
{
   int   slChannelCntT;
   
   
   for(slChannelCntT = 0; slChannelCntT < PCMCIA_CHANNEL_MAX; slChannelCntT++)
   {
      can_unregister_driver(aslCanLogIf[slChannelCntT]);
      aslCanLogIf[slChannelCntT] = -1;
   }

   return(0);
}



//----------------------------------------------------------------------------//
// mc_pcmcia_data_read()                                                      //
// read data from the CAN card, result will be stored in aubDataReadBufG[]    //
//----------------------------------------------------------------------------//
static _U08 mc_pcmcia_data_read(void)
{
   _U08  ubStatusT;
   _U08  ubCountT;
   _U08  ubDelayCntT;


   //--- check if we can read data from the CAN card ----------------
   ubStatusT = mc_pcmcia_status(CP_FIFO_RCV);
   if(ubStatusT == eFIFO_STATUS_EMPTY)
   {
      return(CpErr_FIFO_EMPTY);
   }


   //----------------------------------------------------------------
   // read data from interface
   // access to the CAN interface MUST be slowed down, otherwise
   // some bytes will be set to 0xFF!
   //
   for(ubCountT = 0; ubCountT < PCMCIA_COM_SIZE; ubCountT++)
   {
      aubDataReadBufG[ubCountT] = inb_p(ptsInfoS->port);
      ubStatusT = mc_pcmcia_status(CP_FIFO_RCV);

      //--- check for new data ----------------------------
      if(ubStatusT == eFIFO_STATUS_EMPTY)
      {
         udelay(5);

         ubDelayCntT = 50;
         while(ubDelayCntT)
         {
            //--- delay is 250 us maximum ------------
            ubStatusT = mc_pcmcia_status(CP_FIFO_RCV);
            if(ubStatusT == eFIFO_STATUS_EMPTY) break;
            udelay(5);
            ubDelayCntT--;
         }
      }

   }

   /*
   PK_DBG(  "read success, data=%02X %02X %02X ...", 
            aubDataReadBufG[0], aubDataReadBufG[1], aubDataReadBufG[2]  );
   */
                     
   //--- data read operation was OK ---------------------------------
   return(CpErr_OK);
}


//----------------------------------------------------------------------------//
// CpPcmciaDataWrite()                                                        //
// write data to CAN card, data will be taken from aubDataWriteBufG[]         //
//----------------------------------------------------------------------------//
static _U08 mc_pcmcia_data_write(void)
{
   //--- check if we can write data to the CAN card -----------------
   if(mc_pcmcia_status(CP_FIFO_TRM) == eFIFO_STATUS_FULL)
   {
       return(CpErr_FIFO_FULL);
   }

   //--- write data to card -----------------------------------------
   outsb(ptsInfoS->port, &aubDataWriteBufG[0], PCMCIA_COM_SIZE);  

   return(CpErr_OK);
}



//----------------------------------------------------------------------------//
// mc_pcmcia_reset()                                                          //
// perform reset of the PCMCIA interface                                      //
//----------------------------------------------------------------------------//
int mc_pcmcia_reset(void)
{
   _U32  ulTimeoutValueT;     // timeout value in 'jiffies'
   _U32  ulDelayValueT;       // normalized delay in 'jiffies'
   _U16  uwDelayTimeT;        // delay time in milliseconds


   //----------------------------------------------------------------
   // start the Reset-Sequence
   //
   outb(0x04, (ptsInfoS->port) + 4);
   PK_DBG("Reset seq. 1\n");

   //------------------------------------------------------
   // wait for 200 ms
   // the 'ulDelayValueT' is normalized to jiffie counts
   //
   uwDelayTimeT  = 200;
   ulDelayValueT = (uwDelayTimeT * HZ) / 1000;

   ulTimeoutValueT  = jiffies + ulDelayValueT;
   while(jiffies < ulTimeoutValueT)
   {
      schedule();
   }

   outb(0x80, (ptsInfoS->port) + 4);
   PK_DBG("Reset seq. 2\n");

   ulTimeoutValueT  = jiffies + ulDelayValueT;
   while(jiffies < ulTimeoutValueT)
   {
      schedule();
   }

   outb(0x08, (ptsInfoS->port) + 4);
   PK_DBG("Reset seq. 3\n");

   ulTimeoutValueT  = jiffies + ulDelayValueT;
   while(jiffies < ulTimeoutValueT)
   {
      schedule();
   }

   //----------------------------------------------------------------
   // read data from card and check for "Reset" string
   //
   mc_pcmcia_data_read();

   if( memcmp(&aubDataReadBufG[0], "***[RESET]**", 12) != 0)
   {
      return(CpErr_INIT_FAIL);
   }

   PK_DBG("Reset seq. 4\n");

   //----------------------------------------------------------------
   // clear the card FIFO
   //
   ulTimeoutValueT = 10000;
   while(ulTimeoutValueT)
   {
      if( mc_pcmcia_status(CP_FIFO_RCV) == eFIFO_STATUS_EMPTY) break;
      mc_pcmcia_data_read();
      ulTimeoutValueT--;
   }

   PK_DBG("Reset: %d\n", ulTimeoutValueT);
   return(CpErr_OK);
}


//----------------------------------------------------------------------------//
// mc_pcmcia_set_info()                                                       //
// keep info structure of the PCMCIA interface                                //
//----------------------------------------------------------------------------//
void mc_pcmcia_set_info(struct mc_pcmcia_info_t * ptsInfoV)
{
   ptsInfoS = ptsInfoV;
}


//----------------------------------------------------------------------------//
// mc_pcmcia_status()                                                         //
// retrive the status of the CAN card                                         //
//----------------------------------------------------------------------------//
static _U08 mc_pcmcia_status(_U08 ubFifoTypeV)
{
   _U08  ubFifoStatusT;       // hold the FIFO status


   //----------------------------------------------------------------
   // get the FIFO status 
   //
   ubFifoStatusT = inb( (ptsInfoS->port) + 1);

   //----------------------------------------------------------------
   // check for messages in receive FIFO on the CAN card 
   //
   if(ubFifoTypeV == CP_FIFO_RCV)
   {
      if( !(ubFifoStatusT & FIFO_MASK_RCV_EMPTY) )
      {
         return(eFIFO_STATUS_EMPTY);
      }
      return(eFIFO_STATUS_DATA);
   }

   //----------------------------------------------------------------
   // check for messages in the transmit FIFO on the CAN card
   //
   if(ubFifoTypeV == CP_FIFO_TRM)
   {
      if( !(ubFifoStatusT & FIFO_MASK_TRM_FULL)  )
      {
         return(eFIFO_STATUS_FULL);
      }
      return(eFIFO_STATUS_EMPTY);
   }


   //--- this return should never be reached ------------------------
   return(CpErr_GENERIC);
}


//----------------------------------------------------------------------------//
// mc_pcmcia_task()                                                           //
// check the receive FIFO of the CAN card                                     //
//----------------------------------------------------------------------------//
static void mc_pcmcia_task(unsigned long ulParamV)
{
   register _U08  ubCommandT;       // first byte (command) from PCMCIA card
   register _U16  uwMaxMsgReadT;    // maximum number of messages to read
   _U08           ubDataCntT;       // counter for message data
   _U08           ubDlcT;           // data length code
   _U16           uwStdIdT = 0;     // standard identifier
   _U32           ulExtIdT = 0;     // Extended identifier
   _U16           uwMsgCntT = 1;    // number of messages for queue
   _TsCpCanMsg    tsCanMsgT;        // CAN message structure
   _TsCpPort      tsCanPortT;       // the CAN port
   
   
   //----------------------------------------------------------------
   // check the status of the recieve FIFO
   //
   uwMaxMsgReadT = 32;
   while(uwMaxMsgReadT)
   {
      if(mc_pcmcia_status(CP_FIFO_RCV) != eFIFO_STATUS_DATA) break;
      mc_pcmcia_data_read();
      ubCommandT = aubDataReadBufG[0];
      switch(ubCommandT)
      {
         //-----------------------------------------------------
         // receive Std-ID on CAN 0 / CAN 1
         //
         case 0x70: case 0x71:
            CpMsgClear(&tsCanMsgT);
            uwStdIdT = aubDataReadBufG[2];
            uwStdIdT = uwStdIdT << 8;
            uwStdIdT = uwStdIdT | aubDataReadBufG[1];
            CpMsgSetStdId(&tsCanMsgT, uwStdIdT);

            ubDlcT = aubDataReadBufG[15];
            ubDlcT = ubDlcT >> 4;
            if( ubDlcT > 8)
            {
               ubDlcT = 8;
            }
            CpMsgSetDlc(&tsCanMsgT, ubDlcT);

            for(ubDataCntT = 0; ubDataCntT < ubDlcT; ubDataCntT++)
            {
               CpMsgSetData(&tsCanMsgT, ubDataCntT, aubDataReadBufG[ubDataCntT + 5]);
            }
            
            //--------------------------------------------------
            // the physical channel number is coded in the last
            // bit of the command byte
            //
            tsCanPortT.slLogIf = aslCanLogIf[ubCommandT & 0x01];
            can_queue_rcv_push(&tsCanPortT, &tsCanMsgT, uwMsgCntT);

            break;
         
         //-----------------------------------------------------
         // receive Ext-ID on CAN 0 / CAN 1
         //
         case 0xC0: case 0xC1:
            CpMsgClear(&tsCanMsgT);
            ulExtIdT  = ( aubDataReadBufG[0x04] & 0x1F);
            ulExtIdT  = ulExtIdT << 8;
            ulExtIdT |= aubDataReadBufG[0x03];
            ulExtIdT  = ulExtIdT << 8;
            ulExtIdT |= aubDataReadBufG[0x02];
            ulExtIdT  = ulExtIdT << 8;
            ulExtIdT |= aubDataReadBufG[0x01];
            CpMsgSetExtId(&tsCanMsgT, ulExtIdT);

            ubDlcT = aubDataReadBufG[15];
            ubDlcT = ubDlcT >> 4;
            if( ubDlcT > 8)
            {
               ubDlcT = 8;
            }
            CpMsgSetDlc(&tsCanMsgT, ubDlcT);

            for(ubDataCntT = 0; ubDataCntT < ubDlcT; ubDataCntT++)
            {
               CpMsgSetData(&tsCanMsgT, ubDataCntT, aubDataReadBufG[ubDataCntT + 5]);
            }

            //--------------------------------------------------
            // the physical channel number is coded in the last
            // bit of the command byte
            //
            tsCanPortT.slLogIf = aslCanLogIf[ubCommandT & 0x01];
            can_queue_rcv_push(&tsCanPortT, &tsCanMsgT, uwMsgCntT);
            
            break;
      
      }  
      uwMaxMsgReadT--;
   }
   
   //----------------------------------------------------------------
   // call the write handler for both channel
   //
   tsCanPortT.slPhyIf = 0;
   tsCanPortT.slLogIf = aslCanLogIf[0];
   CpPcmciaMsgWrite(&tsCanPortT, &tsCanMsgT, &uwMsgCntT);

   tsCanPortT.slPhyIf = 1;
   tsCanPortT.slLogIf = aslCanLogIf[1];
   CpPcmciaMsgWrite(&tsCanPortT, &tsCanMsgT, &uwMsgCntT);
      
   //----------------------------------------------------------------
   // do not re-trigger the timer if we want to stop the driver
   if(atomic_read(&ulTimerStopS))
   {
      return;   
   }
   
   init_timer(&tsTimerS);
   tsTimerS.function = mc_pcmcia_task;
   tsTimerS.data     = 0;
   tsTimerS.expires  = jiffies + 2;
   add_timer(&tsTimerS);
   
}

//----------------------------------------------------------------------------//
// mc_pcmcia_task()                                                           //
// check the receive FIFO of the CAN card                                     //
//----------------------------------------------------------------------------//
void mc_pcmcia_task_start(void)
{
   atomic_set(&ulTimerStopS, 0);
   
   init_timer(&tsTimerS);
   tsTimerS.function = mc_pcmcia_task;
   tsTimerS.data     = 0;
   tsTimerS.expires  = jiffies + (2*HZ);
   add_timer(&tsTimerS);
}


//----------------------------------------------------------------------------//
// mc_pcmcia_task()                                                           //
// check the receive FIFO of the CAN card                                     //
//----------------------------------------------------------------------------//
void mc_pcmcia_task_stop(void)
{
   //----------------------------------------------------------------
   // make sure that the timer is not started again
   //
   atomic_set(&ulTimerStopS, 1);
   
   //----------------------------------------------------------------
   // delete a timer that is currently active
   //
   if( timer_pending(&tsTimerS) )
   {
      del_timer_sync(&tsTimerS);
   }
}
