//****************************************************************************//
// File:          canconfig.c                                                 //
// Description:   Test program for writing CAN messages to a socket           //
// Author:        Uwe Koppe                                                   //
// e-mail:        koppe@microcontrol.net                                      //
//                                                                            //
//============================================================================//
// This program is free software; you can redistribute it and/or modify       //
// it under the terms of the GNU Lesser General Public License as published   //
// by the Free Software Foundation; either version 2.1 of the License, or     //
// (at your option) any later version.                                        //
//============================================================================//
//                                                                            //
// Date        History                                                        //
// ----------  -------------------------------------------------------------- //
// 02.01.2004  Initial version                                                //
//                                                                            //
//****************************************************************************//


//------------------------------------------------------------------------------
// CVS version information:
// $Id: canconfig.c,v 1.3 2005/08/15 16:36:33 microcontrol Exp $
//------------------------------------------------------------------------------


/*----------------------------------------------------------------------------*\
** Include files                                                              **
**                                                                            **
\*----------------------------------------------------------------------------*/

#include <sys/socket.h>
#include <signal.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>

#include <can/af_can.h>

#include "can.h"

#include <getopt.h>

#define CAN_PORT_MAX 15

/* Ugh.  But libc5 doesn't provide POSIX types.  */
#include <asm/types.h>

/*----------------------------------------------------------------------------*\
** Definitions                                                                **
**                                                                            **
\*----------------------------------------------------------------------------*/


int  runProg = 1;

//----------------------------------------------------------------------
// quit()
// exit this program gracefully
//----------------------------------------------------------------------
void quit(int signal)
{
   runProg = 0;
}

//----------------------------------------------------------------------
// exit_prog()
// close the filedescriptor and exit this program 
//----------------------------------------------------------------------
void exit_prog(int fd, int err_number) {
   close(fd);
   exit (err_number);
}

//----------------------------------------------------------------------
// usage()
// shows the usage informations
//----------------------------------------------------------------------
void usage(void)
{
   printf("Usage: canconfig <interface> <options>\n\n");
   printf(" interface : can0...canN\n\n");
   printf(" options   : [up|down] [bitrate=<Value> | btr0 bt1 sjw] [ ] \n");
   printf("           : [-?|-h|--help] \n");
}


//----------------------------------------------------------------------
// get_can_state()
// gets state of a socket
//----------------------------------------------------------------------
int get_can_state(int sockfd, int slSockOptT) {
   int opt = -1;
   opt = ioctl(sockfd, SIOC_CAN_GET_STATE, &slSockOptT);
   return opt;
}

//----------------------------------------------------------------------
// connect_can()
// connects a can_port to a socket
//----------------------------------------------------------------------
int connect_can (int can_port, int sockfd) {
   
   struct sockaddr_can can_sock;
   
   can_sock.can_family = AF_CAN;
   can_sock.can_if     = can_port;
   can_sock.can_id     = 0x00FFFFFF;

   if(connect(sockfd, (struct sockaddr*)&can_sock, sizeof(can_sock)) < 0)
   {
      return (1);
   }
   else
      return (0);

}

//----------------------------------------------------------------------
// get_can_nr()
// returns errors or entered user CAN-Number
//----------------------------------------------------------------------
int get_can_nr (char *str_name) {
   int char_index = 3;
   char scan_number[]= "   ";
   if (str_name[0] == 'c' & str_name[1] == 'a' & str_name[2] == 'n') {  
      if (str_name[char_index] == '\0') //the can number is not available
         return -1;
      else {  
         while (str_name[char_index] != '\0') {
            scan_number[char_index-3] = str_name[char_index];  
            char_index++;
         }
      }
   } 
   else {
      // The argument is not adress family name
      return -2;
   }
   return atoi(scan_number);
}

//----------------------------------------------------------------------
// get_can_opt()
// returns a value for an input:
// -1 for Unknown entry
//  1 for "up"
//  2 for "down"
//  3 for "btr0"
//  4 for "btr1"
//  5 for "sjw"
//  6 for "bitrate="
//  7 for help
//----------------------------------------------------------------------
int get_can_opt(char *str_name) {
  
  if (str_name[0] == 'u' & str_name[1] == 'p')
     return 1;
  else if (str_name[0] == 'd' & str_name[1] == 'o' & str_name[2] == 'w' & str_name[3] == 'n')
     return 2;
  else if (str_name[0] == 'b' & str_name[1] == 't' & str_name[2] == 'r' & str_name[3] == '0')
     return 3;
  else if (str_name[0] == 'b' & str_name[1] == 't' & str_name[2] == 'r' & str_name[3] == '1')
     return 4;
  else if (str_name[0] == 's' & str_name[1] == 'j' & str_name[2] == 'w')
     return 5;
  else if (str_name[0] == 'b' & str_name[1] == 'i' & str_name[2] == 't' & str_name[3] == 'r' &
           str_name[4] == 'a' & str_name[5] == 't' & str_name[6] == 'e' & str_name[7] == '=')
     return 6;
  else if ((str_name[0] == '-' & (str_name[1] == 'h' || str_name[1] == '?')) || 
           (str_name[0] == '-' & str_name[1] == '-' & str_name[2] == 'h' & str_name[3] == 'e' &
            str_name[4] == 'l' & str_name[5] == 'p'))
     return 7;
  else return -1;
}

//----------------------------------------------------------------------
// get_bitrate()
// returns an error or entered user bitrate
//----------------------------------------------------------------------
int get_bitrate(char *str_name) {
   char can_bitrate[]= "     ";
   // char_index has the Value 8, because the bitrate should be on this position: "bitrate=XYZ"
   int char_index = 8;
   
   if (str_name[char_index] == '\0')
      return -1;  //No value for bitrate was entered
   else {  
      while (str_name[char_index] != '\0') {
         can_bitrate[char_index-3] = str_name[char_index];  
         char_index++;
      }
      return atoi(can_bitrate);
   }
}

//----------------------------------------------------------------------
// can_info()
// shows all information of a CAN on the screen
//----------------------------------------------------------------------
int can_info (int index, int sockfd, int slSockOptT) {
   printf("can%i\tstate: %i\n", index, ioctl(sockfd, SIOC_CAN_GET_STATE, &slSockOptT));
   printf("\t---Hier wird nun die Information zum can%i angezeigt!---\n", index);
   return 0;
}


int set_can_bitrate(int sockfd, int can_port, int bitrate) {
   int bitrate_value;
   int opt;
   //Muss ergäntzt und korigiert werden!!!!--------------------------------------------------!!!!
   switch (bitrate) {
   case 10 : bitrate_value = 1;
   break;
   case 20 : bitrate_value = 2;
   break;
   case 50 : bitrate_value = 3;
   break;
   case 125 : bitrate_value = 4;
   break;
   case 250 : bitrate_value = 5;
   break;
   case 500 : bitrate_value = 6;
   break;
   case 1000 : bitrate_value = 7;
   break;
   case 101 : bitrate_value = 8;  //
   break;
   case 102 : bitrate_value = 9;  //
   break;
   default : bitrate_value = 0;
   }
   
   opt = ioctl(sockfd, SIOC_CAN_SET_BITRATE, &bitrate_value);
   printf("ioctl(%X) returned the value %d\n", SIOC_CAN_SET_BITRATE, opt);
   return 0;
}

//----------------------------------------------------------------------------//
// main()                                                                     //
//                                                                            //
//----------------------------------------------------------------------------//
int main(int argc, char **argv)
{
   int   opt;
   int   sockfd;
   int   can_port = 0;
   char  buffer[80];
   int   msg_count = 0;
   int   slSockOptT = 125;
   int   level;
   int   bitrate;
   //Option to show all can_ports
   int   opt_a = 0, opt_up=0, opt_down=0, opt_btr0=0, opt_btr1=0, opt_sjw=0, opt_bitrate=0; 
   int   index = 0;
   char  **argv_tmp;
   
   struct can_frame frame;

   // Was erfüllt diese Funktion ???
   signal(SIGINT, quit);

   
   /* Find any entered options. */
   argc--;
   argv++;
   while (argc && (*argv[0] == '-')) {
     if (!strcmp(*argv, "-a"))
        opt_a = 1;
     else if (!strcmp(*argv, "-?") || !strcmp(*argv, "-h") ||
        !strcmp(*argv, "-help") || !strcmp(*argv, "--help")) {
        usage();
        exit (0);
     } else {
        printf(" canconfig: option `%s' not recognised.\n", argv[0]);
        printf(" canconfig: `--help' gives usage information.\n");
        exit (1);
     }
   argv++;
   argc--;
   }
   
   //Open new Socket filedescriptor
   sockfd = socket(AF_CAN, SOCK_RAW, 0);
   if(sockfd < 0)
   {
      printf("Can't open socket AF_CAN\n");
      exit (1);
   }

   // Der Wert für den CAN_PORT_MAX wird noch übergeben!!!
   
   //Show all information of all CAN-Ports
   if ((argc == 0) || (opt_a == 1)) {
      for(index=0; index <= CAN_PORT_MAX; index++) {
         if (connect_can(index,sockfd) == 0) {
            can_info(index,sockfd,slSockOptT);
         } //else printf(" Can not connect to CAN-Port %i\n",index);
      }
      exit_prog(sockfd,0);
      //exit (0);
   }
   
   argv_tmp = argv;
   
   // The next argument is either an address family name, get the can_port number and store in can_port
   if (get_can_nr(*argv_tmp) >= 0) {
      printf("Der eingegeben Can Port ist: %i\n",get_can_nr(*argv_tmp));
      can_port = get_can_nr(*argv_tmp);
   } else if (get_can_nr(*argv_tmp) == -1) {
      printf(" canconfig: the device number for can was not entered.\n");
      printf(" canconfig: `--help' gives usage information.\n");
      exit_prog(sockfd,1);
      //exit (1);
   } else if (get_can_nr(*argv_tmp) == -2) {
      printf(" canconfig: the wrong device name was entered.\n");
      printf(" canconfig: `--help' gives usage information.\n");
      exit_prog(sockfd,1);
      //exit (1);
   }
  
   if (connect_can(can_port,sockfd) == 0) {  
      printf("Successfully conected to CAN-Port %i\n",index);
   } else {
      printf(" Can not connect to CAN-Port %i\n",index);
      exit_prog(sockfd,1);
      //exit(1);
   }
   argc--;
   
   //printf("Anzahl der Parameter: %i\n", argc);
   
   // The next arguments are options for selectet CAN-Port
   // get this options
   if (argc > 0) {
      while (argc > 0) {
         argv_tmp++;
         //printf("hier ist die Option: %i\n", get_can_opt(*argv_tmp));
         switch (get_can_opt(*argv_tmp)) {
         case -1 : 
            printf(" canconfig: you have entered not recognised option.\n");
            printf(" canconfig: `--help' gives usage information.\n");
            exit_prog(sockfd,1);
            //exit(1);         
         break;
         // wenn die Funktion up gewählt ist, dann darf die down Funktion nicht gewählt sein
         case  1 : 
            if (opt_down == 1) {
               printf(" canconfig: you have entered 'up' and 'down' option.\n");
               printf(" canconfig: `--help' gives usage information.\n");
               exit_prog(sockfd,1);
               //exit(1);
            }   
            else {
               opt_up = 1;
               // Hier soll die Funktion aufgerufen werden, die up-Funktion durchführt
            }   
         break;
         
         case  2 : 
            if (opt_up == 1) { 
               printf(" canconfig: you have entered 'up' and 'down' option.\n");
               printf(" canconfig: `--help' gives usage information.\n");
               exit_prog(sockfd,1);
               //exit(1);
            }   
            else {
               opt_down = 1;
               // Hier soll die Funktion aufgerufen werden, die down-Funktion durchführt
            }
         break;
         
         // wenn die Funktion bitrate gewählt ist, dann dürfen die Funktionen 
         // btr0, btr1 und sjw nicht gewählt sein!
         case  3 : 
            if (opt_bitrate == 1) { 
               printf(" canconfig: you have entered wrong combination of bitrate options.\n");
               printf(" canconfig: `--help' gives usage information.\n");
               exit_prog(sockfd,1);
               //exit(1);
            }   
            else 
               opt_btr0 = 1;
         break;

         case  4 : 
            if (opt_bitrate == 1) { 
               printf(" canconfig: you have entered wrong combination of bitrate options.\n");
               printf(" canconfig: `--help' gives usage information.\n");
               exit_prog(sockfd,1);
               //exit(1);
            }  
            else 
               opt_btr1 = 1;
         break;         
         
         case  5 : 
            if (opt_bitrate == 1) { 
               printf(" canconfig: you have entered wrong combination of bitrate options.\n");
               printf(" canconfig: `--help' gives usage information.\n");
               exit_prog(sockfd,1);
               //exit(1);
            }  
            else 
               opt_sjw = 1;
         break;
         
         case  6 : 
            if ((opt_btr0 == 1) || (opt_btr1 == 1) || (opt_sjw == 1)) { 
               printf(" canconfig: you have entered wrong combination of bitrate options.\n");
               printf(" canconfig: `--help' gives usage information.\n");
               exit_prog(sockfd,1);
               //exit(1);
            }  
            else {
               opt_bitrate = 1;
               bitrate = get_bitrate(*argv_tmp);
               printf("Hie ist eure gewählte Bitrate: %i\n",bitrate);
               // Hier soll die Funktion aufgerufen werden, die die neue bitrate festlegt
               set_can_bitrate(sockfd,can_port,bitrate);
            }
         break;
         
         case 7 :
           usage();
           exit_prog(sockfd,0);
         break;
         }
         argc--;
      }
      // Wenn alle drei Funktionen btr0, btr1 und sjw eingegeben wurden, 
      // so sollen nun die einstellungen übermittelt werden
      if (opt_sjw & opt_btr1 & opt_btr0) {
        // Hier soll nun die Funktion aufgerufen werden, die diese Einstellungen übermittelt
      }
   } else // No options were entered, so show information for selected CAN-Port
     if (connect_can(can_port,sockfd) == 0) {  
        can_info(can_port,sockfd,slSockOptT);         
     } //else printf(" Can not connect to CAN-Port %i\n",index);

   

   //----------------------------------------------------------
   // set socket options
   //
   //opt = ioctl(sockfd, SIOC_CAN_SET_BITRATE, &slSockOptT);
   //printf("ioctl(%X) returned the value %d\n", SIOC_CAN_SET_BITRATE, opt);
   
   //opt = getsockopt(sockfd, SOL_SOCKET, 1, &slSockOptT, sizeof(slSockOptT));
   //printf("getsockopt() returned the value %d\n", opt);
   
   exit_prog(sockfd,0);
//   close(sockfd);
   
}
